(*
                  ufrmMainForm
       (c) 2003 Paul C. L. Willmott, VP9MU

 This program is free software; you can redistribute it
 and/or modify it under the terms of the GNU General
 Public License as published by the Free Software
 Foundation; either version 2 of the License, or at
 your option, any later version.

 This program is distributed in the hope that it will
 be useful, but WITHOUT ANY WARRANTY; without even the
 implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE.  See the GNU General Public
 License for more details.

 You should have received a copy of the GNU General
 Public License along with this program; if not, write
 to the Free Software Foundation, Inc., 59 Temple
 Place, Suite 330, Boston, MA  02111-1307  USA

            Contact : vp9mu@amsat.org
*)

(*
     Revision History
     ----------------

     1.0.0      Created July 21, 2003 PCLW
*)


unit ufrmMainForm;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Menus, StdCtrls, umdlBPSKDet, umdlBPSK400, ExtCtrls, ComCtrls , umdlAO40FECCodec ;


type
  TfrmMainForm = class(TForm)
    mnuMainMenu: TMainMenu;
    mnuFile: TMenuItem;
    mnuExit: TMenuItem;
    mnuTools: TMenuItem;
    mnuHelp: TMenuItem;
    mnuAbout: TMenuItem;
    mnuOptions: TMenuItem;
    imgHidden: TImage;
    Panel1: TPanel;
    imgRawBlock: TImage;
    Panel2: TPanel;
    imgWaveForm: TImage;
    pgcntrlTabs: TPageControl;
    tbshtStatus: TTabSheet;
    tbshtLogging: TTabSheet;
    btnStop: TButton;
    btnStart: TButton;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    lblRegGood: TLabel;
    lblRegBad: TLabel;
    lblRegTotal: TLabel;
    lblFECGood: TLabel;
    lblFECBad: TLabel;
    lblFECTotal: TLabel;
    Label6: TLabel;
    lblCmdGood: TLabel;
    lblCMDBad: TLabel;
    lblCMDTotal: TLabel;
    procedure mnuAboutClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure mnuExitClick(Sender: TObject);
    procedure mnuOptionsClick(Sender: TObject);
    procedure btnStartClick(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
  private
    { Private declarations }
    tblTranslate : array[MIN_FREQ..MAX_FREQ] of integer ;
    k3 : double ;
    iRegGood : integer ;
    iRegBad : integer ;
    iFECGood : integer ;
    iFECBad : integer ;
    iCMDGood : integer ;
    iLastTime : cardinal ;
    iLastStatus : integer ;
    pLastBlock : TABlock ;
    bBlockWaiting : boolean ;
    procedure FFTDataReady(var msg:TMessage) ; message MSG_FFTDATARDY ;
    procedure StatusChange(var msg:TMessage) ; message MSG_STATUSCHANGE ;
    procedure DataReady(var msg:TMessage) ; message MSG_DATARDY ;
    procedure WriteBlock ;
  public
    { Public declarations }
  end;

var
  frmMainForm: TfrmMainForm;

implementation

{$R *.dfm}

uses

     ufrmAbout,
     umdlOptions,
     ufrmOptions,
     umdlFFT,
     umdlIOCntrl,
     Math ;

procedure TfrmMainForm.mnuAboutClick(Sender: TObject);
var
     i : integer ;
begin
     frmAbout := TfrmAbout.Create(Self) ;
     frmAbout.ShowModal ;
     StopReceive ;
     i := Self.Handle ;
     i := StartSoundReceive(i,giAudioInputDevice) ;
end;

procedure TfrmMainForm.FormCreate(Sender: TObject);
var
     i : integer ;
begin

     Top := GetSetting(ksINIFile,Name,'Top',Top) ;
     Left := GetSetting(ksINIFile,Name,'Left',Left) ;
     giAudioInputDevice  := GetSetting(ksINIFile,'Options','Audio In Device',0) ;
     giAudioOutputDevice := GetSetting(ksINIFile,'Options','Audio Out Device',0) ;
     gsLogPath := GetSetting(ksINIFile,'Options','Log Path','c:\') ;
     gsCallsign := GetSetting(ksINIFile,'Options','Callsign','callsign') ;
     for i := MIN_FREQ to MAX_FREQ do begin
         tblTranslate[i] := (i-MIN_FREQ)*imgWaveForm.Width div (MAX_FREQ-MIN_FREQ) ;
     end ;
     k3 := imgWaveForm.Height / 2.0 ;

     imgHidden.Width := imgWaveForm.Width ;
     imgHidden.Height := imgWaveForm.Height ;

     i := Self.Handle ;
     i := StartSoundReceive(i,giAudioInputDevice) ;
     btnStart.Enabled := i <> 0 ;
     btnStop.Enabled := not btnStart.Enabled ;
     iRegGood := 0 ;
     iRegBad  := 0 ;
     iFECGood := 0 ;
     iFECBad  := 0 ;
     iCMDGood := 0 ;
     iLastTime := 0 ;
     iLastStatus := -1 ;
     bBlockWaiting := false ;

end;

procedure TfrmMainForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
     StopReceive ;
     if (bBlockWaiting) then begin
         WriteBlock ;
         bBlockWaiting := false ;
     end ;
     SaveSetting(ksINIFile,Name,'Top',Top) ;
     SaveSetting(ksINIFile,Name,'Left',Left) ;
     SaveSetting(ksINIFile,'Options','Audio In Device',giAudioInputDevice) ;
     SaveSetting(ksINIFile,'Options','Audio Out Device',giAudioOutputDevice) ;
     SaveSetting(ksINIFile,'Options','Log Path',gsLogPath) ;
     SaveSetting(ksINIFile,'Options','Callsign',gsCallsign) ;
     Action := caFree ;
end;

procedure TfrmMainForm.mnuExitClick(Sender: TObject);
begin
     Close ;
end;

procedure TfrmMainForm.mnuOptionsClick(Sender: TObject);
begin
     frmOptions := TfrmOptions.Create(Self) ;
     frmOptions.ShowModal ;
end;

procedure TfrmMainForm.btnStartClick(Sender: TObject);
var
     i : integer ;
begin
     i := Self.Handle ;
     i := StartSoundReceive(i,giAudioInputDevice) ;
     btnStart.Enabled := i <> 0 ;
     btnStop.Enabled := not btnStart.Enabled ;
end;

procedure TfrmMainForm.DataReady(var msg: TMessage);

var

     _pBlock : TDataBlock ;
     pBlock : TABlock absolute _pBlock ;
     cc : char ;

procedure DisplayBlock(var pBlock:TABlock) ;

var

     ii, xx, yy, iWidth, iHeight : integer ;
     pRect : TRect ;
     iTime : Cardinal ;
begin

//   Clear Previous Block

     imgRawBlock.Canvas.Brush.Color := clBtnFace ;
     imgRawBlock.Canvas.FillRect(ClientRect);
     imgRawBlock.Canvas.Font.Name := 'Courier New' ;
     imgRawBlock.Canvas.Font.Size := 8 ;
     imgRawBlock.Canvas.Font.Style := [] ;

     iWidth := imgRawBlock.Width div 64 ;
     iHeight := imgRawBlock.Height div 8 ;

//   Process All Bytes

     for ii := 0 to SizeOf(TABlock)-1 do begin

         xx := ii mod 64 ;
         yy := ii div 64 ;

         pRect.Top    := yy * iHeight ;
         pRect.Left   := xx * iWidth ;
         pRect.Bottom := (yy + 1) * iHeight -1 ;
         pRect.Right  := (xx + 1) * iWidth - 1 ;

         if (pBlock.pDataX[ii] and $80 <>0) then begin
             imgRawBlock.Canvas.Brush.Color := clBlack ;
             imgRawBlock.Canvas.FillRect(pRect) ;
             imgRawBlock.Canvas.Font.Color := clWhite ;
         end
         else begin
             imgRawBlock.Canvas.Brush.Color := clBtnFace ;
             imgRawBlock.Canvas.FillRect(pRect) ;
             imgRawBlock.Canvas.Font.Color := clBlack ;
         end ;
         cc := char(pBlock.pDataX[ii] and $7f) ;
         if (cc<' ') then begin
             cc := ' ' ;
         end ;
         imgRawBlock.Canvas.TextOut(pRect.Left,pRect.Top,cc) ;

     end ;

     iTime := pBlock.pDataX[$1AD] ;
     iTime := iTime shl 8 or pBlock.pDataX[$1AC] ;
     iTime := iTime shl 8 or pBlock.pDataX[$1AB] ;
     iTime := iTime shl 8 or pBlock.pDataX[$1AA] ;
     iTime := iTime shl 8 or pBlock.pDataX[$1A9] ;
     iTime := iTime shl 8 or pBlock.pDataX[$1A8] ;

     case msg.WParam of
     0 : begin // Regular CRCC OK
             if (bBlockWaiting) then begin
                 WriteBlock ;
             end ;
             pLastBlock := pBlock ;
             WriteBlock ;
             bBlockWaiting := false ;
             Inc(iRegGood) ;
             lblRegGood.Caption := IntToStr(iRegGood) ;
             Inc(iCmdGood) ;
             lblCmdGood.Caption := IntToStr(iCmdGood) ;
         end ;
     1 : begin // Regular CRCC Bad
             if (bBlockWaiting) then begin
                 WriteBlock ;
             end ;
             pLastBlock := pBlock ;
             bBlockWaiting := true ;
             Inc(iRegBad) ;
             lblRegBad.Caption := IntToStr(iRegBad) ;
         end ;
     2 : begin // FEC RS Success
             Inc(iFECGood) ;
             lblFECGood.Caption := IntToStr(iFECGood) ;
             if (iTime<>iLastTime) or (iLastStatus<>0) then begin
                 Inc(iCmdGood) ;
                 lblCmdGood.Caption := IntToStr(iCmdGood) ;
                 pLastBlock := pBlock ;
                 WriteBlock ;
                 bBlockWaiting := false ;
             end ;
         end ;
     3 : begin // FEC RS Failure
             Inc(iFECBad) ;
             lblFECBad.Caption := IntToStr(iFECBad) ;
             if (bBlockWaiting) then begin
                 pLastBlock := pBlock ;
                 WriteBlock ;
                 bBlockWaiting := false ;
             end ;
         end ;
     end ;
     lblRegTotal.Caption := IntToStr(iRegGood+iRegBad) ;
     lblFECTotal.Caption := IntToStr(iFECGood+iFECBad) ;
     lblCmdTotal.Caption := IntToStr(iCmdGood) ;

     iLastTime := iTime ;
     iLastStatus := msg.WParam ;

end { DisplayBlock } ;

begin
//
//application.MessageBox('Data Ready','',MB_OK) ;
     GetDataBlock(_pBlock) ;
     DisplayBlock(pBlock) ;
end;

procedure TfrmMainForm.FFTDataReady(var msg: TMessage);
var
     pBuffer : TFFTOutBuf ;
     var i : integer ;
begin
     if (IOCntrl.GetFFTData(pBuffer,MIN_FREQ,MAX_FREQ)) then begin
         imgHidden.Canvas.Brush.Color := clBlack ;
         imgHidden.Canvas.FillRect(ClientRect);
         imgHidden.Canvas.MoveTo(0, imgHidden.Height * (255-pBuffer[0]) div 255  );
         imgHidden.Canvas.Pen.Color := clGreen ;
         for i := MIN_FREQ+1 to MAX_FREQ do begin
             imgHidden.Canvas.LineTo(tblTranslate[i], imgHidden.Height * (255-pBuffer[i]) div 255) ;
         end ;
         imgWaveForm.Picture := imgHidden.Picture ;
     end ;
end;

procedure TfrmMainForm.StatusChange(var msg: TMessage);
var
     sMsg : string ;
begin
//
//application.MessageBox('Status Change','',MB_OK) ;
case msg.WParam of
DLLSTAT_OFF : sMsg := 'Off' ;
DLLSTAT_RX_NO_SIGNAL : sMsg := 'No Signal' ;
DLLSTAT_RX_SYNC_SEARCH : sMsg := 'Sync Search' ;
DLLSTAT_RX_BLOCK_READ : sMsg := 'Block Read' ;
DLLSTAT_CPUSLOW : sMsg := 'CPU Slow' ;
end ;

end;

procedure TfrmMainForm.btnStopClick(Sender: TObject);
begin
     StopReceive ;
     btnStop.Enabled := false ;
     btnStart.Enabled := not btnStop.Enabled ;
end;


procedure TfrmMainForm.WriteBlock;
var
     fOutFile : file of TABlock ;
     sOutFile : string ;
begin
     sOutFile := FormatDateTime('YYMMDD',Now) ;
     sOutFile := ExcludeTrailingBackslash(gsLogPath)+'\T'+sOutFile+'@'+gsCallsign+'.raw' ;
     {$I-}
     AssignFile(fOutFile,sOutFile) ;
     FileMode := fmOpenReadWrite ;
     Reset(fOutFile) ;
     {$I+}
     if (IOResult <> 0) then begin
         Rewrite(fOutFile) ;
     end ;
     if (IOResult = 0) then begin
         Seek(fOutFile, FileSize(fOutFile)) ;
         Write(fOutFile,pLastBlock) ;
     end ;
     CloseFile(fOutFile) ;

end { WriteFile } ;

end.
