(*
                  umdlBPSKDet
       (c) 2003 Paul C. L. Willmott, VP9MU

          Based on code from AO40RCV
         (c) 2000 Moe Wheatley AE4JY
             ae4jy@mindspring.com

 This program is free software; you can redistribute it
 and/or modify it under the terms of the GNU General
 Public License as published by the Free Software
 Foundation; either version 2 of the License, or at
 your option, any later version.

 This program is distributed in the hope that it will
 be useful, but WITHOUT ANY WARRANTY; without even the
 implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE.  See the GNU General Public
 License for more details.

 You should have received a copy of the GNU General
 Public License along with this program; if not, write
 to the Free Software Foundation, Inc., 59 Temple
 Place, Suite 330, Boston, MA  02111-1307  USA

            Contact : vp9mu@amsat.org
*)

(*
     Revision History
     ----------------

     1.0.0      Created July 21, 2003 PCLW
*)

unit umdlBPSKDet;

interface

uses

   umdlErrorCodes,
   umdlAO40FECCodec,
   umdlLPIIR,
   umdlBPSKDetTbl,
   umdlSound,
   Messages,
   Windows ;

const

   BLOCK_SIZE = 514 ;

type

   TVectData = array[0..16-1] of integer ;
   TSyncData = array[0..10-1] of integer ;
   TDataBlock = packed array[0..BLOCK_SIZE-1] of byte ;

   TBPSKDet = class(TObject)
   private
      bTryFreqSearchAgain : boolean ;

// other classes
      pBwIIR560 : TLPIIR ;
//variables
      bFreqAdjustOK : boolean ;
      bCRCok : boolean ;
      bBitSample : boolean ;
      iBitAcc0 : cardinal ;
      iBitAcc1 : cardinal ;
      iBitPhzErrCnt : integer ;
      iDecodeState : integer ;
      iBitCnt : integer ;
      iByteCnt : integer ;
      pIQPhaseArray : TVectData ;
      iIQPhzIndex : integer ;
      iLastPkPos : integer ;
      iBlockSize : integer ;
      iRxFrequency : integer ;
      iFirState : integer ;
      dPhaseError : double ;
      dI0 : double ;		// 4 stage I/Q delay line variables
      dI1 : double ;
      dI2 : double ;
      dI3 : double ;
      dQ0 : double ;
      dQ1 : double ;
      dQ2 : double ;
      dQ3 : double ;
      dBitPhaseInc : double ;
      dBitPhasePos : double ;
      pSyncAve : array[0..12-1] of double ;
      dNCOphzinc : double ;
      dSampleFreq : double ;
      pQue : array[0..LPFIR_LEN-1] of TComplex ;
      pFECCodec : TAO40FECCodec ;
      pDataBuf : TDataBlock ;


// Local variables for various functions that need to be saved between calls
      dPhzErrAve : double ;
      dVcoPhz : double ;
      iPkPos : integer ;
      iNewPkPos : integer ;
      iBitPos : integer ;
      hWnd : HWND ;

//methods
      procedure DecodeSymb( newsamp : TComplex);
      function SymbSync(sample:TComplex) : boolean ;
      procedure CalcPhaseError(IQ : TComplex);
      function CalcCRC(var buf:TDataBlock) : word ;

   public

   //public variables

      function GetBPSKSymb : boolean ;

      procedure Init(iFs:integer;BlockSize:integer;Handle:HWND) ;
      function ProcBPSKDet(var pIn:TFFTWaveBuffer) : integer ;
      function GetRXFrequency : integer ;
      procedure GetVectorData(var pVectData : TVectData) ;
      procedure GetSyncData(var pSyncData:TSyncData);
      procedure ResetDetector ;
      function OkToCheckFreq : boolean ;
      procedure SetRXFrequency(iFreq:integer;bManual:boolean);
      function GetDataBlock(var pData:TDataBlock):boolean ;
      procedure FECBlockReady(iStatus:integer) ;
      constructor Create ;
      destructor Destroy ;
   end ;

var

   pBPSKDet : TBPSKDet ; // 1 instance

implementation

uses

     umdlFFT ,
     umdlBPSK400 ;

const

   STATE_IDLE_SEARCH = DLLSTAT_RX_NO_SIGNAL ;
   STATE_SYNC_SEARCH = DLLSTAT_RX_SYNC_SEARCH ;
   STATE_BLOCK_FILL = DLLSTAT_RX_BLOCK_READ ;

//   define some constants
     PHZ_180_BMIN =	0.0       ; // 0
     PHZ_180_BMAX = Pi	      ; // Pi
     PHZ_0_BMIN   = Pi	      ; // Pi
     PHZ_0_BMAX   = 2.0 * Pi  ; // 2Pi
//
     PHZ_180_QMIN = Pi / 4.0       ; //  Pi/4
     PHZ_180_QMAX = Pi * 3.0 / 4.0 ; // 3Pi/4
     PHZ_0_QMIN   = Pi * 5.0 / 4.0 ; // 5Pi/4
     PHZ_0_QMAX   = Pi * 7.0 / 4.0 ; // 7Pi/4

     PI2 = 2.0 * Pi ;

//   define the BPSK symbols

     SYM_NOCHANGE = 0 ; //Stay the same phase
     SYM_P90      = 1 ; //Plus 90  deg
     SYM_P180     = 2 ; //Plus 180 deg
     SYM_M90      = 3 ; //Minus 90 deg

     SYMB_RATE  = 800.0 ;
     TSYM       = 1.0 / SYMB_RATE + 0.0 ; // TSYM == symbol period
     BYTE_SYNC  = $50505050 ;
     BLOCK_SYNC = $3915ED30 ;
     SYNC_LIMIT = 700 ;


AMSAT_CYC2_TBL : array[0..255] of word = (
	$0000, $1021, $2042, $3063, $4084, $50a5, $60c6, $70e7, $8108, $9129, $a14a, $b16b, $c18c, $d1ad, $e1ce, $f1ef,
	$1231, $0210, $3273, $2252, $52b5, $4294, $72f7, $62d6, $9339, $8318, $b37b, $a35a, $d3bd, $c39c, $f3ff, $e3de,
	$2462, $3443, $0420, $1401, $64e6, $74c7, $44a4, $5485, $a56a, $b54b, $8528, $9509, $e5ee, $f5cf, $c5ac, $d58d,
	$3653, $2672, $1611, $0630, $76d7, $66f6, $5695, $46b4, $b75b, $a77a, $9719, $8738, $f7df, $e7fe, $d79d, $c7bc,
	$48c4, $58e5, $6886, $78a7, $0840, $1861, $2802, $3823, $c9cc, $d9ed, $e98e, $f9af, $8948, $9969, $a90a, $b92b,
	$5af5, $4ad4, $7ab7, $6a96, $1a71, $0a50, $3a33, $2a12, $dbfd, $cbdc, $fbbf, $eb9e, $9b79, $8b58, $bb3b, $ab1a,
	$6ca6, $7c87, $4ce4, $5cc5, $2c22, $3c03, $0c60, $1c41, $edae, $fd8f, $cdec, $ddcd, $ad2a, $bd0b, $8d68, $9d49,
	$7e97, $6eb6, $5ed5, $4ef4, $3e13, $2e32, $1e51, $0e70, $ff9f, $efbe, $dfdd, $cffc, $bf1b, $af3a, $9f59, $8f78,
	$9188, $81a9, $b1ca, $a1eb, $d10c, $c12d, $f14e, $e16f, $1080, $00a1, $30c2, $20e3, $5004, $4025, $7046, $6067,
	$83b9, $9398, $a3fb, $b3da, $c33d, $d31c, $e37f, $f35e, $02b1, $1290, $22f3, $32d2, $4235, $5214, $6277, $7256,
	$b5ea, $a5cb, $95a8, $8589, $f56e, $e54f, $d52c, $c50d, $34e2, $24c3, $14a0, $0481, $7466, $6447, $5424, $4405,
	$a7db, $b7fa, $8799, $97b8, $e75f, $f77e, $c71d, $d73c, $26d3, $36f2, $0691, $16b0, $6657, $7676, $4615, $5634,
	$d94c, $c96d, $f90e, $e92f, $99c8, $89e9, $b98a, $a9ab, $5844, $4865, $7806, $6827, $18c0, $08e1, $3882, $28a3,
	$cb7d, $db5c, $eb3f, $fb1e, $8bf9, $9bd8, $abbb, $bb9a, $4a75, $5a54, $6a37, $7a16, $0af1, $1ad0, $2ab3, $3a92,
	$fd2e, $ed0f, $dd6c, $cd4d, $bdaa, $ad8b, $9de8, $8dc9, $7c26, $6c07, $5c64, $4c45, $3ca2, $2c83, $1ce0, $0cc1,
	$ef1f, $ff3e, $cf5d, $df7c, $af9b, $bfba, $8fd9, $9ff8, $6e17, $7e36, $4e55, $5e74, $2e93, $3eb2, $0ed1, $1ef0
) ;

{ TBPSKDet }

function TBPSKDet.CalcCRC(var buf:TDataBlock) : word ;

// Returns the 16 bit CRC of the specified data buffer.

var

     crc : cardinal ;
     i : integer ;

begin
     crc := $FFFF ;
     for i := 0 to BLOCK_SIZE-1 do begin
         crc := (crc shl 8) xor AMSAT_CYC2_TBL[((crc shr 8) xor buf[i]) and $00FF] ;
     end ;
     result := crc and $ffff ;
end { TBPSKDet.CalcCRC } ;

procedure TBPSKDet.CalcPhaseError(IQ: TComplex);

const

     PLP_K = 250 ;
     P_GN = (0.0002) ;
     I_GN = 6e-6 ;
     ERR_K = (1.0/14.09) ;	//make error into Hz.
var
     samp : double ; // static

begin

     if (IQ.y > 0.0) then begin
         samp := -IQ.x ;
     end
     else begin
         samp := IQ.x ;
     end ;

     dPhzErrAve := (1.0-1.0/PLP_K)*dPhzErrAve + ((1.0*ERR_K)/PLP_K)*samp ;
     if ((dPhzErrAve > 5.0) or (dPhzErrAve < -5.0 )) then begin
         dNCOphzinc := dNCOphzinc + (dPhzErrAve*I_GN) ;
         if (dNCOphzinc < 0.157) then begin
             dNCOphzinc := 0.157 ;
         end
         else begin
             if (dNCOphzinc > 2.67) then begin
                 dNCOphzinc := 2.67 ;
             end ;
         end ;
     end ;

     dPhaseError := dPhzErrAve * P_GN ;

end;

constructor TBPSKDet.Create;
begin
     iRxFrequency := 1600 ;
     dSampleFreq := WAVE_SAMPLE_RATE ;
     iIQPhzIndex := 0 ;
     dPhaseError := 0.0 ;
     dNCOphzinc := PI2 * iRxFrequency / dSampleFreq ;
     dBitPhaseInc := 1.0/8000.0 ; //bit oversampling period
     bBitSample := true ;
     pBwIIR560 := TLPIIR.Create ;
     pFECCodec := TAO40FECCodec.Create ;
end;

procedure TBPSKDet.DecodeSymb(newsamp:TComplex) ;
//////////////////////////////////////////////////////////////////////
//  Decode the new symbol sample
//////////////////////////////////////////////////////////////////////
var
     vect, nvect : TComplex ;
     energy : double ;
     bit : boolean ;
     iTmp : integer ;
begin
     dI3 := dI2 ;
     dQ3 := dQ2 ;
     dI2 := dI1 ;
     dQ2 := dQ1 ;
     dI1 := dI0 ;
     dQ1 := dQ0 ;
     dI0 := newsamp.x ;
     dQ0 := newsamp.y ;
     vect.y := (dI2*dI0 + dQ2*dQ0)+(dI3*dI1 + dQ3*dQ1) ;
     vect.x := (dI2*dQ0 - dQ2*dI0)+(dI3*dQ1 - dQ3*dI1) ;

     if (bBitSample) then begin //process every other bit time
         bit := (vect.y < 0.0) ; { ***** }
         iBitAcc1 := iBitAcc1 shl 1 ; //store phase 1 bits
         if (bit) then begin
             iBitAcc1 := iBitAcc1 or 1 ;
         end ;
         Inc(iBitCnt) ;

         if (iBitCnt >= 8) then begin

             iBitCnt := 0 ;

             for iTmp := 0 to 7 do begin
                 if ((iBitAcc1 shl iTmp) and $80 = $80) then begin
                     pFECCodec.AddSymbol($f0) ;
                 end
                 else begin
                     pFECCodec.AddSymbol($10) ;
                 end ;
             end ;

         end ;

//   calculate normalized phase vectors for display use

         energy := Sqrt(vect.x*vect.x + vect.y*vect.y) / 1.0E3 ;

         if (energy > 10.0) then begin
             nvect.x := (vect.x/energy) ;
             nvect.y := (vect.y/energy) ;
             CalcPhaseError(nvect) ;
             pIQPhaseArray[iIQPhzIndex] := Trunc(nvect.x) ;
             Inc(iIQPhzIndex) ;
             pIQPhaseArray[iIQPhzIndex] := Trunc(nvect.y) ;
             Inc(iIQPhzIndex) ;
         end
         else begin
             pIQPhaseArray[iIQPhzIndex] := 2 ;
             Inc(iIQPhzIndex) ;
             pIQPhaseArray[iIQPhzIndex] := 2 ;
             Inc(iIQPhzIndex) ;
         end ;

         iIQPhzIndex := iIQPhzIndex and $000F ; //mod 16 index

         case iDecodeState of

             STATE_IDLE_SEARCH : begin
                 bFreqAdjustOK := true ;
                 if (iBitAcc0 = BYTE_SYNC) then begin	//if phase 0 stream is correct
                                          						//then flip to use it instead
                     iBitCnt := 0 ;
                     iByteCnt := 0 ;
                     bBitSample := not bBitSample ;
                     iDecodeState := STATE_SYNC_SEARCH ;
                 end ;
                 if (iBitAcc1 = BYTE_SYNC) then begin //stick with phase 1 stream
                     iBitCnt := 0 ;
                     iByteCnt := 0 ;
                     iDecodeState := STATE_SYNC_SEARCH ;
                 end ;
             end ; { STATE_IDLE_SEARCH }

             STATE_SYNC_SEARCH : begin
                 bFreqAdjustOK := false ;
                 if (iBitAcc1 = BYTE_SYNC) then begin
                     iBitCnt := 0 ;
                     iByteCnt := 0 ;
                 end
                 else begin
                     if (iBitCnt = 0) then begin
                         Inc(iByteCnt) ;
                         if (iByteCnt >= SYNC_LIMIT) then begin // if haven't gotten blk sync
                             iByteCnt := 0 ;                    // timeout and try again
                             iDecodeState := STATE_IDLE_SEARCH ;
                         end ;
                         if (iBitAcc1 = BLOCK_SYNC) then begin
                             iByteCnt := 0 ;
                             iDecodeState := STATE_BLOCK_FILL ;
                         end ;
                     end ;
                 end ;
             end ; { STATE_SYNC_SEARCH }

             STATE_BLOCK_FILL : begin
                 if (iBitCnt = 0) then begin
                     if (iByteCnt < BLOCK_SIZE) then begin
                         pDataBuf[iByteCnt] := iBitAcc1 and $FF ;
                         Inc(iByteCnt) ;
                     end
                     else begin
                         if (CalcCRC(pDataBuf) = 0) then begin
                             bCRCok := true ;
                         end
                         else begin
                             bCRCok := false ;
                             pDataBuf[0] := pDataBuf[0] or $80 ;
                             bTryFreqSearchAgain := true ;
                             iDecodeState := STATE_IDLE_SEARCH ;
                         end ;
                         iDecodeState := STATE_SYNC_SEARCH ;
                         if (bCRCok) then begin
                             iTmp := 0 ;
                         end
                         else begin
                             iTmp := 1 ;
                         end ;
                         PostMessage(hWnd, MSG_DATARDY, iTmp, 0) ;
                         iByteCnt := 0 ;
                     end ;
                 end ;
             end ; { STATE_BLOCK_FILL }

         end ; // case

     end
     else begin
         bit := (vect.y < 0.0) ;
         iBitAcc0 := iBitAcc0 shl 1 ;		//store phase 0 bits
         if (bit) then begin
             iBitAcc0 := iBitAcc0 or 1 ;
//             pFECCodec.AddSymbol($10) ; // Hard-Limiting Solution
         end
         else begin
//             pFECCodec.AddSymbol($F0) ; // Hard-Limiting Solution
         end ;
     end ;

     bBitSample := not bBitSample ;

end { TBPSKDet.DecodeSymb } ;

destructor TBPSKDet.Destroy;
begin
     inherited ;
     pBwIIR560.Destroy ;
     pFECCodec.Destroy ;
end;

procedure TBPSKDet.FECBlockReady(iStatus:integer) ;
var
     pBlock : TDataBlock ;
     _pBlock : TABlock absolute pBlock ;
     iTmp : integer ;
begin
     _pBlock := pFECCodec.ReconABlock ;
     pDataBuf := pBlock ;
     PostMessage(hWnd, MSG_DATARDY, iStatus, 0) ;
end;

function TBPSKDet.GetBPSKSymb: boolean;
begin
      // ?????
end;

function TBPSKDet.GetDataBlock(var pData: TDataBlock): boolean;
//////////////////////////////////////////////////////////////////////
// Called to get the data block
//////////////////////////////////////////////////////////////////////
begin
     pData := pDataBuf ;
     result := bCRCok ;
end ;

function TBPSKDet.GetRXFrequency: integer;
begin
     result := iRxFrequency ;
end;

procedure TBPSKDet.GetSyncData(var pSyncData: TSyncData);
//////////////////////////////////////////////////////////////////////
// Called to get the symbol vectors
//////////////////////////////////////////////////////////////////////
var
     i : integer ;
begin
     for i := 0 to 10 -1 do begin
         pSyncData[i] := 500 ;
     end ;
end;

procedure TBPSKDet.GetVectorData(var pVectData:TVectData) ;
//////////////////////////////////////////////////////////////////////
// Called to get the symbol vectors
//////////////////////////////////////////////////////////////////////
var
     i : integer ;
begin
     for i := 0 to 16 -1 do begin
         pVectData[i] := pIQPhaseArray[i] ;
     end ;
end;

procedure TBPSKDet.Init(iFs, BlockSize: integer;Handle:HWND);

//       Initialize BpskDet buffers and pointers

var

     i : integer ;

begin

     dSampleFreq := iFs ;
     hWnd := Handle ;
     pFECCodec.hWnd := Handle ;
     iBlockSize := BlockSize ; //size data input buffer

     ResetDetector ;

     dBitPhasePos := 0.0 ;
     iBitAcc0 := 0 ;
     iBitAcc1 := 0 ;
     iBitCnt := 0 ;
     iByteCnt := 0 ;
     iDecodeState := STATE_IDLE_SEARCH ;
     bFreqAdjustOK := true ;
     iBitPhzErrCnt := 0 ;
// Init a bunch of "static" variables used in various member fuctions
     dVcoPhz := 0.0 ;
     iPkPos := 0 ;
     iNewPkPos := 0 ;
     iBitPos := 0 ;
     dI3 := 0.0 ;
     dI2 := 0.0 ;
     dI1 := 0.0 ;
     dI0 := 0.0 ;
     dQ3 := 0.0 ;
     dQ2 := 0.0 ;
     dQ1 := 0.0 ;
     dQ0 := 0.0 ;
     dPhzErrAve := 0.0 ;
     iLastPkPos := 0 ;
     bCRCok := false ;
     for i := 0 to LPFIR_LEN-1 do begin
         pQue[i].x := 0.0 ;
         pQue[i].y := 0.0 ;
     end ;
     iFirState := LPFIR_LEN-1 ;
     pBwIIR560.Init(IIR_LPCOEF560) ;
     bTryFreqSearchAgain := false ;
end;

function TBPSKDet.OkToCheckFreq:boolean ;
begin
     result :=  bFreqAdjustOK or bTryFreqSearchAgain ;
end;

function TBPSKDet.ProcBPSKDet(var pIn:TFFTWaveBuffer) : integer ;
//////////////////////////////////////////////////////////////////////
// Main routine called to process the next block of data 'pIn'.
//////////////////////////////////////////////////////////////////////
var

     vcophz : double ;
     i,j : integer ;
     acc : TComplex ;
     Kptr : integer ; // index into LPfir_COEF
     Firptr : integer ; // index into pQue

begin

     vcophz := dVcoPhz ; // Maybe Static issue?

     bTryFreqSearchAgain := false ;
     for i := 0 to iBlockSize-1 do begin // put new samples into Queue

//   Generate complex sample by mixing input sample with NCO's sin/cos

         vcophz := vcophz + dNCOphzinc + dPhaseError ;
         if (vcophz > PI2) then begin //handle 2 Pi wrap around
             vcophz := vcophz - PI2 ;
         end ;
         acc.x := pIn.pBuffer[i] * Cos(vcophz) ;
         acc.y := pIn.pBuffer[i] * Sin(vcophz) ;
//   FIR LP filter
         pQue[iFirState].x := acc.x ; //place in circular Queue
         pQue[iFirState].y := acc.y ;
         Firptr := 0 ; // pQue ;
         Kptr := {LPfir_COEF + } LPFIR_LEN - iFirState ;
         acc.x := 0.0 ;
         acc.y := 0.0 ;
         for j := 0 to LPFIR_LEN-1 do begin
              acc.x := acc.x + pQue[Firptr].x * LPfir_COEF[Kptr] ;
              acc.y := acc.y + pQue[Firptr].y * LPfir_COEF[Kptr] ;
              Inc(Firptr) ;
              Inc(Kptr) ;
         end ;
         Dec(iFirState) ;
         if (iFirState < 0) then begin
             iFirState := LPFIR_LEN-1 ;
         end ;
         acc := pBwIIR560.CalcFilter(acc) ; // a gentler, kinder IIR filter
//   Bit Timing synchronization
         if (SymbSync(acc)) then begin
            DecodeSymb(acc) ;
         end ;
     end ;

     dVcoPhz := vcophz ;
     iRxFrequency := Trunc((0.5+((dNCOphzinc + dPhaseError)*dSampleFreq/PI2 ) )) ;

     result := iDecodeState ;

end;

procedure TBPSKDet.ResetDetector;
//////////////////////////////////////////////////////////////////////
// Called to reset the detector
//////////////////////////////////////////////////////////////////////
var
     i : integer ;
begin

     for i := 0 to LPFIR_LEN-1 do begin
         pQue[i].x := 0.0 ; // fill delay buffer with zero
         pQue[i].y := 0.0 ;
     end ;

     for i := 0 to 16-1 do begin
         pIQPhaseArray[i] := 1 ;
     end ;

     for i := 0 to 12-1 do begin
         pSyncAve[i] := 0.0 ; // initialize the array
     end ;

end;

procedure TBPSKDet.SetRXFrequency(iFreq: integer; bManual: boolean);
//////////////////////////////////////////////////////////////////////
// Called to change the Rx frequency
// freq is the new center frequency
// manual is TRUE if it is a user wanting to over ride
// Setidle == TRUE forces the decoder back to the idle search mode
//////////////////////////////////////////////////////////////////////
begin
     if (bManual or (OkToCheckFreq and (iFreq>0))) then begin
         iDecodeState := STATE_IDLE_SEARCH ;
         if (iFreq <> iRxFrequency) then begin
             iRxFrequency := iFreq ;
             dNCOphzinc := PI2* iFreq/dSampleFreq ;
             dPhzErrAve := 0.0 ;
         end ;
     end ;
end ;

function TBPSKDet.SymbSync(sample:TComplex): boolean ;
//////////////////////////////////////////////////////////////////////
// Called at Fs/16 rate to calculate the symbol sync position
// Returns TRUE if at center of symbol.
// Sums up the energy at each sample time, averages it, and picks the
//   sample time with the highest energy content.
//////////////////////////////////////////////////////////////////////
var

     Trigger : boolean ;
     max : double ;
     energy : double ;
     BitPos : integer ;
     i : integer ;

begin

     BitPos := iBitPos ;

     energy := (sample.x*sample.x) + (sample.y*sample.y) ;
     pSyncAve[BitPos] := (1.0-1.0/75.0)*pSyncAve[BitPos] + (1.0/75.0)*energy ;

     if (BitPos = iPkPos) then begin // see if at middle of symbol
         Trigger := true ;
     end
     else begin
         Trigger := false ;
     end ;

     if (BitPos = HALF_TBL[iPkPos])	then begin //don't change pk pos until
         iPkPos := iNewPkPos ; // halfway into next bit.
     end ;
     Inc(BitPos) ;
     if (BitPos>=10) then begin
         BitPos := 0 ;
     end ;
     dBitPhasePos := dBitPhasePos + dBitPhaseInc ;
     if (dBitPhasePos >= TSYM) then begin
//   here every symbol time
         dBitPhasePos := dBitPhasePos - TSYM ; //keep phase bounded
         BitPos := 0 ;
         max := -1e10 ;
         for i := 0 to 10-1 do begin //find maximum energy pk
             energy := pSyncAve[i] ;
             if (energy>max) then begin
                 iNewPkPos := i ;
                 max := energy ;
             end ;
         end ;
     end ;
     iBitPos := BitPos ;

     result := Trigger ;
end ;

(*
                 Processing AO-40 Data at the Symbol level

                         by James Miller G3RUH

                              2003 Jun 25

These notes describe methods of processing the 400 symbols/sec generated by an
ADC attached to the integrate-and-dump of a G3RUH P3 400 bps PSK Decoder.

Symbols could also be generated by a computer's soundcard DSP process, or indeed
by simulation.

Familiarity with the AMSAT P3 telemetry idiom is assumed.  Some sample code is
given.

For details of the AO-40 satellite telemetry see:

 http://www.amsat-dl.org/p3d/pubtelem.zip

For details of the experimental FEC telemetry encoding format see:

 http://www.ka9q.net/ao40/


Outline
-------
The tasks of the service program are:

 1. Service the RS232 port
 2. Condition the data to signed format and scale
 3. Differentially decode successive symbols
 4. Store the data stream in a circular buffer
 5. Scan the data for Sync vector/Output frame of data.

Analogue to Digital Converter - note
------------------------------------
The 8-bit ADC measures the voltage at the PSK decoder integrate-and-dump at a
rate of 400/s.  It outputs its byte RS-232 style at 9600 baud.  Each of these
bytes represents a "symbol".

The bytes represent:
                            ADC
  Symbol                   Value
  ------------------------------
  Strongest "1"            0xFF
  Weak 0/1  (erasure)      0x80 [*]
  Strongest "0"            0x00
  ------------------------------

* Note: This level should be 0x80, but is not guaranteed to be so, as it
depends on the user installing the ADC with the correct voltage divider
resistors, and on the equipment's PSU voltage.  In principle things can be
calibrated out by software.


1. Service the RS232 port
   ----------------------
The RS-232 port should be set to 9600 baud, format 8N1 or 8N2.  Bytes will
arrive at 400/s, so this procedure, and all subsequent processing needs to be
efficient.

2. Condition the data to signed format and scale
   ---------------------------------------------
Denoting the byte value read as v%, process thus:

    v = v%-adc_mean
    v =  v*adc_scale

The NOMINAL value of adc_mean is 128.  The value is constant, though in a
specific installation it may have to be established by a calibration procedure
t.b.d. and perhaps retained in a set-up file.

The NOMINAL value of adc_scale is 1.0 but could well be higher or lower
depending on the audio level the user inputs to the Demodulator hardware. It is
a constant that may have to be set up by a calibration procedure t.b.d. and
perhaps retained in a set-up file.

Note that v and adc_scale are numbers with finer than integer resolution, and
will need scaling accordingly if you do not wish to use floating point maths.

At this stage, on full amplitude signals, the data should be a number within
the range approx -120.0  +120.0 and should not exceed +/-127.  It doesn't
matter if this range, with good signals, is as little as (say) +/-80.

3. Differentially decode successive symbols
   ----------------------------------------
The raw data was differentially encoded at the spacecraft; a "1" is represented
by a change in channel polarity, and a "0" by no change.   Thus, perform:

   sym= -(v*vo )         /* differential decode ( note "-" sign) */
   symbol%=sym/128       /* rescale */
   vo=v                  /* preserve v */

The last operation saves the current value of  v  for use next time around.  It
should be initialised to 0.

The rescale operation is signed, and may be implemented by an arithmetic
shift.  symbol% is a signed byte.


4. Store the data stream in a circular buffer
   ------------------------------------------
An 8192 byte circular buffer is suggested, initialised throughout to 0x80.  The
pointer should be initialised to 0.

symbol% should be converted to unsigned and stored in the buffer:

   ip_buf[ptr] = symbol%+128

Increment the pointer AFTER the next stage.


5. Scan the data for Sync vector/Output frame of data
   --------------------------------------------------
The structure of this stage depends on whether we are dealing with FEC data
frames or normal uncoded frames.  This is because the FEC frames have an
embedded sync vector, whereas normal uncoded frames have a preceding sync
vector.

  FEC Frames (5200 symbols)
  -------------------------
  Each time a new symbol has been stored, look back through the preceding 5200
  symbols and perform a correlation operation with every 80th symbol against the
  local sync vector replica.

  If the sync gain exceeds a threshold, declare sync_65 found.

  Output the 5200 symbol frame (to the Viterbi/RS decoder package)

  Normal Uncoded frames (4096 symbols + 16 crcc)
  ----------------------------------------------
  Each time a new symbol has been stored, look back through the preceding 31
  symbols and perform a correlation operation with the 31 symbols against the
  local sync vector replica (PN31).

  If the sync gain exceeds a threshold, declare sync_31 found.

  When 514*8 further symbols have been read, the frame has been acquired. Slice
  the symbols to top bit only, pack them 8 to a byte, do a crcc check, mark the
  MSB of byte 0 and output the 512 byte block.

  Examples of both these procedures is appended.  They can easily be combined
  into one package.


Sync Detection - note
---------------------
Sync detection is performed by correlating symbols against a local replica.
The basic operation is:

   total =  sum{ s(i)*S(i) }

 where
   sum{} is the sum of the product indicated for i=1 .. N where N = 31 or 65
   s(i) are the N (signed) symbols being tested
   S(i) are the N (signed) local sync vector replica elements; valued +1 or -1

The total must be  normalised, and for this we require an estimate of the
magnitude of the signal symbols s():

   mag = sqrt(sum{ s(i)*s(i)/N })

 where
   sqrt is the square root operation.

Dividing these, we have a measure of the process gain, thus:

                  sum{ s(i)*S(i) }
   sync_gain = -----------------------
               sqrt(sum{ s(i)*s(i)/N })

If the symbols represent noise, or data uncorrelated with the sync vector,
sync_gain will be a random variable with zero mean.

However, if the symbols are the sync vector, their energy will combine
coherently and the sync_gain will take a value of order +N.

By choosing a threshold above the highest expected values generated due to
noise etc [about 5*sqrt(N)], the presence of a sync vector can be reliably
detected.

Example Coding
--------------
The language is BBCBASIC

Variables are either:
 4-byte integers           e.g. ptr%
 floating point variables, e.g. adc_mean
 strings                   e.g. sync$

Arrays: DIM sync% 65  defines a byte array named sync% of 65 bytes

Array access:  array?i  returns the contents of array[i]
               array?i=x writes byte x to array[i]

Hex numbers are written e.g. &12EF

Procedures are introduced with DEF PROCxxxx and closed with ENDPROC.



+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

Example of FEC Data collection
==============================
:
PROCinit_RS423
:
REM Set up PN65 sync vector  bit pattern
DIM sync% 65
sync$="11111110000111011110010110010010000001000100110001011101011011000"
FOR i%=1 TO 65:  b$=MID$(sync$,i%,1):  sync%?(i%-1)=VAL(b$): NEXT : REM Q&D
:
ip_ptr%=0      : REM Circular input buffer pointer
vo=0           : REM Initialise "old" i/p value
sync_found%=0  : REM gets set to -1 when sync detected.
adc_mean =128  : REM Note: need a method to get these two real-time calibrated
adc_scale=1.5  : REM to get raw data to have decent span, say 0x10 to 0xF0
:
FOR i%=0 TO 8191: ip_buf%?i%=&80: NEXT : REM flush input buffer
:
REM Main loop; runs forever
REPEAT
  REPEAT
    v%=FNserial(1,port%,0,0)  : REM /very/ crude handler!
  UNTIL v%<>-1             : REM Waiting for a byte from ADC on serial port
  :
  REM new symbol at last ... remember, 400 calls/sec at this point
  v = v%-adc_mean          : REM it's a byte, and "zero" is ~128
  v = v*adc_scale          : REM rescale range to +/- max
  sym%=-(v*vo )>>7         : REM Diff decode
  vo = v                   : REM Preserve old raw voltage
  ip_buf%?ip_ptr%=sym%+128 : REM New sym to buffer;
                             REM it's a byte; 0 to 127 = "0"
                             REM            128 to 255 = "1"; ~128 is "poor"
  PROCtest_sync65(ip_ptr%)
  IF sync_found% THEN
      REM ... we've collected 5200 symbols for the FEC block, so save to file.
    PROCdump_block_5200
    sync_found%=0
  ENDIF
  ip_ptr%+=1                : REM update circular ip buffer pointer mod 8192
  ip_ptr%=ip_ptr% AND &1FFF
UNTIL 0                     : REM Forever
END

REM------------------------------------
REM. SUB-ROUTINES for main loop
REM------------------------------------
:
DEF PROCtest_sync65(ip_ptr%)
sync_volts  =0
frame_energy=1 : REM Mustn't be zero or we're open to a divide-by-zero
                 REM   error on "silence" ;-)
:
ptr%=(ip_ptr%-5200) AND &1FFF: REM Start search here
FOR i%=0 TO 64
  sync_bit%=sync%?i%
  sym%=ip_buf%?ptr% - 128  : REM -128 to -1 = "0"; 0 to 127 = "1";  ~0 is "poor"
  IF sync_bit% THEN sync_volts+=sym% ELSE  sync_volts-=sym%
  frame_energy+=sym%*sym%
  ptr%=(ptr%+80) AND &1FFF : REM Pointer to next sync bit position, MOD 8192
NEXT
sync_gain=sync_volts/SQR(frame_energy/65)
IF sync_gain>41 THEN sync_found%=-1
ENDPROC
:
REM Looks back over previous 5200 symbols and see whether sync vector is
REM present.
REM Max sync_gain = 65 on perfectly uniform, noise free signal
REM
REM On noise only, the spread of sync_gain 'seen' is about +/-30 (~4 sigma),
REM although 35.5 will be seen if you wait long enough
REM On noise, likelihood of sync_gain > 40 is remote (~1/day)
REM
REM On-air experiments show that if  sync_gain < 41 is detected, it is
REM extremely unlikely that the associated block will FEC decode.
REM
REM So set threshold at 41.  [Evaluated 2003 Jun 19]

REM ----------------------------------------------------------------------------
:
DEF PROCdump_block_5200
FOR i%=0 TO 5200-1
  ptr%=(ip_ptr%-5200+i%) AND &1FFF
  sym%=ip_buf%?ptr%   : REM get symbol from the symbol buffer
  op_buf%?i%=sym%     : REM Copy the symbol to the o/p buffer
NEXT
REM At this point the FEC block is at op_buf% ready for the Viterbi/RS
REM processes, though here we simply dump it to disc.
:
fh%=OPENOUT "feclog"
PTR#fh%=EXT#fh%
SYS "OS_GBPB",2,fh%,op_buf%,5200  : REM Append latest block
CLOSE#fh%
ENDPROC

+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+



Example of Normal Uncoded frame Data collection
===============================================
DIM ip_buf% 8192  : REM Circular buffer for input stream
DIM op_buf% 514*8 : REM Output buffer for block (bits)
DIM op_blk% 514   : REM Output buffer for block (bytes)
:
PROCinit_RS423

REM Set up PN31 sync vector (0x3915ED30) bit pattern
DIM sync% 32:  sync$="0011100100010101111011010011000"
FOR i%=1 TO 31:  b$=MID$(sync$,i%,1):  sync%?(i%-1)=VAL(b$): NEXT
:
ip_ptr%=0      : REM Circular input buffer pointer
vo=0           : REM Initialise "old" i/p value
sync_found%=0  : REM gets set to -1 when sync detected.
adc_mean =128  : REM Note: need a method to get these two real-time calibrated
adc_scale=1.5  : REM to get raw data to have decent span, say 0x10 to 0xF0
:
FOR i%=0 TO 8191: ip_buf%?i%=&80: NEXT : REM flush input buffer
:
REM Main loop; runs forever
REM Main loop; runs forever
REPEAT
  REPEAT
    v%=FNserial(1,port%,0,0)  : REM /very/ crude handler!
  UNTIL v%<>-1             : REM Waiting for a byte from ADC on serial port
  :
  REM new symbol at last ... remember, 400 calls/sec at this point
  v = v%-adc_mean          : REM it's a byte, and "zero" is ~128
  v = v*adc_scale          : REM rescale range to +/- max
  sym%=-(v*vo )>>7         : REM Diff decode
  vo = v                   : REM Preserve old raw voltage
  ip_buf%?ip_ptr%=sym%+128 : REM New sym to buffer;
                             REM it's a byte; 0 to 127 = "0"
                             REM            128 to 255 = "1"; ~128 is "poor"
  IF sync_found% THEN
      REM ... we're collecting bits for the block
    op_buf%?block_bit_count%=sym%+128
    block_bit_count%+=1
    IF block_bit_count%=514*8 THEN PROCdump_block_512: sync_found%=0
  ELSE
    REM ... we're waiting for sync vector of new block
    PROCtest_sync31(ip_ptr%)
    block_bit_count%=0
  ENDIF
  ip_ptr%+=1                : REM update circular ip buffer pointer mod 8192
  ip_ptr%=ip_ptr% AND &1FFF
UNTIL 0                     : REM Forever
END



REM------------------------------------
REM. SUB-ROUTINES for main loop
REM------------------------------------
:
DEF PROCtest_sync31(ip_ptr%)
sync_volts=0
frame_energy=1 : REM Mustn't be zero or we're open to a divide-by-zero
                 REM   error on "silence" ;-)
:
ptr%=(ip_ptr% - 31) AND &1FFF: REM Start search here
FOR i%=0 TO 30
  sync_bit%=sync%?i%
  sym%=ip_buf%?ptr% - 128  : REM -128 to -1 = "0";  0-127  = "1"; ~0 is "poor"
  IF sync_bit% THEN sync_volts+=sym% ELSE  sync_volts-=sym%
  frame_energy+=sym%*sym%
  ptr%=(ptr%+ 1) AND &1FFF : REM Pointer to next sync bit position, MOD 8192
NEXT
sync_gain=sync_volts/SQR(frame_energy/31)
IF sync_gain>25 THEN sync_found%=TRUE
ENDPROC
REM Look at latest 31 bits and see whether sync vector is present.
REM Max sync_gain = 31 on perfectly uniform, noise free signal
REM On noise only, the spread of sync_gain is typ. +/-19 , ~3.4  sigma level
REM
REM On signal, if sync_gain < 29, crcc OK is very unlikely
REM
REM Threshold of about 25 ensures some poor blocks for Pauls' bit-merge system

REM ----------------------------------------------------------------------------

DEF PROCdump_block_512
crcc%=&FFFF
FOR i%=0 TO 513*8 STEP 8
  byte%=0
  FOR j%=0 TO 7
    sym%=op_buf%?(i%+j%)     : REM get sym from the sym buffer
    bit%=(sym%AND&80)>>7     : REM Use MSbit only (Viterbi would use full value)
    byte%=(byte%<<1) OR bit% : REM Build up the byte
    PROCbit_crcc             : REM Cycle crc division using bit
  NEXT
  op_blk%?(i%>>3)=byte%      : REM Plant the new byte in o/p block
NEXT
IF (crcc% AND &FFFF) THEN op_blk%?0=op_blk%?0 OR &80 : REM crc failed, set MSB
:
REM At this point new block is at  op_blk%  and is ready for telemetry display
REM program.  Here, we'll just dump it to disc for now.
:
fh%=OPENUP "tlmlog"
PTR#fh%=EXT#fh%
SYS"OS_GBPB",2,fh%,op_blk%,512  : REM Append latest block
CLOSE#fh%
ENDPROC

REM ----------------------------------------------------------------------------
DEF PROCbit_crcc
test% = (bit%<<15 EOR crcc%) AND &8000
crcc% = crcc% << 1
IF test% THEN crcc% = crcc% EOR &1021
ENDPROC


*)

end.
