(*
                  umdlBPSKDetTbl
       (c) 2003 Paul C. L. Willmott, VP9MU

          Based on code from AO40RCV
         (c) 2000 Moe Wheatley AE4JY
             ae4jy@mindspring.com

 This program is free software; you can redistribute it
 and/or modify it under the terms of the GNU General
 Public License as published by the Free Software
 Foundation; either version 2 of the License, or at
 your option, any later version.

 This program is distributed in the hope that it will
 be useful, but WITHOUT ANY WARRANTY; without even the
 implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE.  See the GNU General Public
 License for more details.

 You should have received a copy of the GNU General
 Public License along with this program; if not, write
 to the Free Software Foundation, Inc., 59 Temple
 Place, Suite 330, Boston, MA  02111-1307  USA

            Contact : vp9mu@amsat.org
*)

(*
     Revision History
     ----------------

     1.0.0      Created July 21, 2003 PCLW
*)

unit umdlBPSKDetTbl;

interface

uses

     SysUtils,
     umdlLPIIR ;
     
const

     PI2 = 2.0 * Pi ;

(*
//////////////////////////////////////////////////////////////////////
//  IIR 3rd order Butterworth LP filter coeficients.
//       The order is as follows:
//   B1, B0, A1   for the 1st order section
/*==================================================================*/
/*                                                                  */
/*                          -1                                      */
/*                 B0 + B1 z                                        */
/*          H(z) = -----------                                      */
/*                         -1                                       */
/*                  1 + A1 z                                        */
/*                                                                  */
/*  The block diagram used in the implementation is given below:    */
/*                                                                  */
/*          input               w(n)   B0         output            */
/*            -----> + ----------+-----[>--> + ----->               */
/*                   |           |           |                      */
/*                   |        +--+--+        |                      */
/*                   |        |Delay|        |                      */
/*                   |        +--+--+        |                      */
/*                   |   -A1     |w(n-1) B1  |                      */
/*                   +----<]-----+-------[>--+                      */
/*                                                                  */
/*==================================================================*/
//   B2, B1, B0, A2, A1  for the 2nd order section
//
/*                                                                  */
/*                          -1       -2                             */
/*                 B0 + B1 z   + B2 z                               */
/*          H(z) = --------------------                             */
/*                         -1       -2                              */
/*                 1 + A1 z   + A2 z                                */
/*                                                                  */
/*  The block diagram used in the implementation is given below:    */
/*                                                                  */
/*          input               w(n)   B0         output            */
/*            -----> + ----------+-----[>--> + ----->               */
/*                   |           |           |                      */
/*                   |        +--+--+        |                      */
/*                   |        |Delay|        |                      */
/*                   |        +--+--+        |                      */
/*                   |   -A1     |w(n-1) B1  |                      */
/*                   +----<]-----+-------[>--+                      */
/*                   |           |           |                      */
/*                   |        +--+--+        |                      */
/*                   |        |Delay|        |                      */
/*                   |        +--+--+        |                      */
/*                   |   -A2     |w(n-2) B2  |                      */
/*                   +----<]-----+-------[>--+                      */
/*                                                                  */
/*==================================================================*/
//////////////////////////////////////////////////////////////////////

*)

IIR_LPCOEF560 : TLPCoef = (
// 560 Hz BW (Fs=8000)
	0.18280856, 0.18280856, -0.63438288,
	0.03928822, 0.07857644, 0.03928822, 0.64874732, -1.49159444
) ;

IIR_LPCOEF75 : TLPCoef = (
// 75 Hz BW (Fs=8000)
	0.04372781, 0.04372781, -0.91254438,
		0.00199557, 0.00399113, 0.00199557, 0.9127189, -1.90473663
) ;

IIR_LPCOEF5 : TLPCoef = (
// 5 Hz BW (Fs=8000)
	0.00366072, 0.00366072, -0.99267855,
		1.34499461e-5, 2.68998923e-5, 1.34499461e-5, 0.99267865, -1.99262485
);
IIR_LPCOEF25 : TLPCoef = (
// 25 Hz BW (Fs=8000)
	0.0, 0.0, -0.0,
		9.52844435E-5, 1.90568887E-4, 9.52844435E-5, 0.97258189, -1.97220075
);


//bit1

LPFIR_LEN = 21 ;

LPfir_COEF : array[0..LPFIR_LEN*2-1] of double = (
 -0.00039171975,
  -0.0035452392,
   -0.010823429,
   -0.021086277,
   -0.027422866,
   -0.018390751,
    0.016085453,
    0.076403237,
     0.14856133,
     0.20793765,
     0.23097343,
     0.20793765,
     0.14856133,
    0.076403237,
    0.016085453,
   -0.018390751,
   -0.027422866,
   -0.021086277,
   -0.010823429,
  -0.0035452392,
 -0.00039171975,
//
 -0.00039171975,
  -0.0035452392,
   -0.010823429,
   -0.021086277,
   -0.027422866,
   -0.018390751,
    0.016085453,
    0.076403237,
     0.14856133,
     0.20793765,
     0.23097343,
     0.20793765,
     0.14856133,
    0.076403237,
    0.016085453,
   -0.018390751,
   -0.027422866,
   -0.021086277,
   -0.010823429,
  -0.0035452392,
 -0.00039171975
);
// bit sync half distance table
// index any position 0 to 9 and it returns a position halfway from the index
HALF_TBL : array[0..10-1] of integer = (
   5,	// 0
   6,	// 1
   7,	// 2
   8,	// 3
   9,	// 4
   0,	// 5
   1,	// 6
   2,	// 7
   3,	// 8
   4	// 9
);

implementation

end.
