(*
                   umdlFFT
       (c) 2003 Paul C. L. Willmott, VP9MU

          Based on code from AO40RCV
         (c) 2000 Moe Wheatley AE4JY
             ae4jy@mindspring.com

      Which was based on a slightly modified
            version of Takuya OOURA's
          original radix 4 FFT package.
      Copyright(C) 1996-1998 Takuya OOURA
            ooura@mmm.t.u-tokyo.ac.jp

 This program is free software; you can redistribute it
 and/or modify it under the terms of the GNU General
 Public License as published by the Free Software
 Foundation; either version 2 of the License, or at
 your option, any later version.

 This program is distributed in the hope that it will
 be useful, but WITHOUT ANY WARRANTY; without even the
 implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE.  See the GNU General Public
 License for more details.

 You should have received a copy of the GNU General
 Public License along with this program; if not, write
 to the Free Software Foundation, Inc., 59 Temple
 Place, Suite 330, Boston, MA  02111-1307  USA

            Contact : vp9mu@amsat.org
*)

(*
     Revision History
     ----------------

     1.0.0      Created July 21, 2003 PCLW
*)

unit umdlFFT;

interface

const

   WAVE_SAMPLE_RATE = 8000 ; // Hz
   WAVE_SAMPLE_SIZE = 2048 ; // number of data chunks to process at a time from input
   FFT_SIZE         = WAVE_SAMPLE_SIZE div 2 ;
   SQRT_FFT_SIZE    = 32 ;   // sqrt(1024)
   K_2PI            = 2.0 * PI ; // (8.0*atan(1))
   FREQ_SCALE       = WAVE_SAMPLE_RATE / FFT_SIZE * 1.0 ;

type


   TFFTBuffer = packed array[0..FFT_SIZE-1] of double ;
   TSinCosTbl = packed array[0..FFT_SIZE div 2 -1] of double ;
   TWorkArea  = packed array[0..SQRT_FFT_SIZE+2-1] of integer ;
   TFFTOutBuf = array[0..FFT_SIZE-1] of integer ;

   TFFT = class(TObject)
   private
      bOverload    : boolean    ;
      iLastAve     : integer    ;
      dGain        : double     ;
      dClip        : double     ;
      bLogMode     : boolean    ;
      bLastLogMode : boolean    ;
      iAveSize     : integer    ;
      SinCosTbl    : TSinCosTbl ;
      WindowTbl    : TFFTBuffer ;
      pFFTAveBuf   : TFFTBuffer ; // Even index's hold average
      pFFTInBuf    : TFFTBuffer ;
      WorkArea     : TWorkArea  ;

      procedure makewt(nw:integer; var ip:TWorkArea; var w:TSinCosTbl;wip:integer { index into w } ) ;
      procedure makect(nc:integer; var ip:TWorkArea; var c:TSinCosTbl;cip:integer { index into c } ) ;
      procedure bitrv2(n :integer; var ip:TWorkArea; ipp:integer {index into ip} ;var _a;aip:integer {index into a});
      procedure cftfsub(n:integer; var a:TFFTBuffer; var w:TSinCosTbl) ;
      procedure rftfsub(n:integer; var a:TFFTBuffer; nc:integer; var c:TSinCosTbl;cpp:integer { index into c}) ;
      procedure cft1st(n :integer; var a:TFFTBuffer; var w:TSinCosTbl) ;
      procedure cftmdl(n, l :integer; var a : TFFTBuffer; var w:TSinCosTbl) ;

   public
      procedure CalcFFT(var InBuf:TFFTBuffer) ;
      procedure ResetFFT ;
      function GetFFTData(iStart, iStop : integer;var pOutBuf:TFFTOutBuf) : boolean ;
      function GetFFTPeak(iCenter, iRange:integer) : integer ;
      procedure SetFFTParams(iAve:integer; Gain:double; iType :integer);
      function CalcCorrelation : integer ;
      constructor Create ;
   end ;

implementation

uses

   Math ;

var

   CoreData : array[0..1000] of double ;

//////////////////////////////////////////////////////////////////////
// A pure input sin wave ... Asin(wt)... will produce an fft output
//   peak of (N*A/4)^2  where N is FFT_SIZE.
// To convert to a Power dB range:
//   PdBmax = 10*log10( (N*A/4)^2 + K_C ) + K_B
//   PdBmin = 10*log10( 0 + K_C ) + K_B
//  if (N*A/4)^2 >> K_C
//  Then K_B = PdBmax - 20*log10( N*A/4 )
//       K_C = 10 ^ ( (PdBmin-K_B)/10 )
//  for power range of 0 to 100 dB with input(A) of 32767 and N=2048
//			K_B = -44.494132  and K_C = 2.81458e4
// To eliminate the multiply by 10, divide by 10 so for an output
//		range of 0 to 100dB the stored value is 0.0 to 10.0
//   so final constant K_B = -4.4494132

const

K_B = -3.8473533 ;
K_C = 7.036445e3 ;
K_ROOT = (1.0/4.0) ;
K_ROOTGN = 289.626 ;		//(A*N/8)^(2/4) / 10

CORR_SIZE = 200 ;

// correlation data obtained experimentally from an "idle" sync
// stream of BPSK telemetry data.  Used for frequency search
// by correlating this data with incomming data and finding
// peak.

CORREL_ARRAY : array[0..CORR_SIZE-1] of double = (
   0.362245,
   0.726316,
   1.66759,
   3.16434,
   3.23513,
   1.96128,
   0.70784,
   0.374534,
   0.729382,
   3.95711,
   4.50391,
   3.84693,
   0.651327,
   0.797754,
   1.6695,
   3.64964,
   4.77086,
   4.60388,
   2.8685,
   1.43263,
   0.619755,
   0.577281,
   0.896443,
   0.461981,
   0.760545,
   0.949337,
   1.41347,
   2.52696,
   5.13609,
   5.94787,
   5.54369,
   3.00362,
   2.60419,
   3.56092,
   5.17626,
   6.68988,
   6.75372,
   5.46975,
   3.68561,
   2.6888,
   3.24726,
   6.82705,
   7.36939,
   6.70891,
   3.00662,
   3.44309,
   4.47693,
   6.4349,
   7.55567,
   7.38642,
   5.6472,
   4.16071,
   3.89477,
   5.25489,
   7.34024,
   7.63627,
   6.6901,
   4.44909,
   3.43585,
   4.03928,
   6.66361,
   7.47145,
   7.06346,
   4.5308,
   3.16513,
   3.62696,
   5.25152,
   6.76131,
   6.82186,
   5.53309,
   3.7598,
   2.76923,
   2.10084,
   1.54187,
   1.24132,
   1.23556,
   1.8578,
   2.51438,
   3.53985,
   5.49941,
   6.61482,
   6.44274,
   4.69096,
   3.16871,
   3.09956,
   4.51231,
   6.59608,
   6.88844,
   5.93795,
   3.7122,
   2.63779,
   3.13365,
   5.77138,
   6.57505,
   6.16347,
   3.62135,
   2.29918,
   1.53204,
   1.40262,
   1.68062,
   0.952041,
   0.628609,
   0.654283,
   1.32375,
   2.57326,
   6.05034,
   6.58526,
   5.91831,
   2.36128,
   2.76495,
   3.77691,
   5.74282,
   6.85329,
   6.67771,
   4.91943,
   3.40543,
   2.87214,
   4.28405,
   6.35763,
   6.64636,
   5.69119,
   3.46939,
   2.41756,
   1.66494,
   1.20856,
   1.13422,
   1.5627,
   2.00414,
   2.71538,
   3.63634,
   5.26686,
   6.76103,
   6.81503,
   5.51441,
   3.7451,
   2.74146,
   3.47253,
   6.94726,
   7.47843,
   6.80747,
   3.27749,
   3.512,
   4.51889,
   6.49195,
   7.59787,
   7.41865,
   5.65572,
   4.1754,
   3.868,
   5.2456,
   5.2456,
   7.584,
   6.62401,
   4.4005,
   3.28556,
   3.85461,
   6.55589,
   7.35226,
   6.93346,
   4.35185,
   3.0462,
   3.55776,
   5.20049,
   6.68993,
   6.74035,
   5.43445,
   3.67545,
   2.69616,
   2.1245,
   5.42449,
   5.95119,
   5.27591,
   1.80686,
   0.841571,
   0.476186,
   0.506915,
   0.404722,
   0.483894,
   0.622903,
   0.487344,
   1.16776,
   2.46524,
   4.51152,
   4.79418,
   3.82918,
   1.66203,
   0.713153,
   0.936349,
   3.69663,
   4.48424,
   4.06287,
   1.52113,
   0.542617,
   0.489631,
   1.69747,
   3.16875,
   3.21413,
   1.89019,
   0.593202,
   0.363721
) ;

{ TFFT }

procedure TFFT.bitrv2(n :integer; var ip:TWorkArea; ipp:integer {index into ip} ;var _a;aip:integer {index into a});

var

     a : TFFTBuffer absolute _a ;

     j, j1, k, k1, l, m, m2 : integer ;
     xr, xi : double ;

begin

     ip[0 + ipp] := 0 ;
     l := n ;
     m := 1 ;
     while ((m shl 2) < l) do begin
         l := l shr 1 ;
         for j := 0 to m-1 do begin
             ip[m + j + ipp] := ip[j + ipp] + l ;
         end ;
         m := m shl 1 ;
     end ;
     if ((m shl 2) > l) then begin
         for k := 1 to m-1 do begin
             for j := 0 to k-1 do begin
                j1 := (j shl 1) + ip[k + ipp];
                k1 := (k shl 1) + ip[j + ipp];
                xr := a[j1 + aip] ;
                xi := a[j1 + 1 + aip] ;
                a[j1 + aip] := a[k1 + aip] ;
                a[j1 + 1 + aip] := a[k1 + 1 + aip] ;
                a[k1 + aip] := xr ;
                a[k1 + 1 + aip] := xi ;
             end ;
         end ;
     end
     else begin
         m2 := m shl 1 ;
         for k := 1 to m-1 do begin
             for j := 0 to k-1 do begin
                 j1 := (j shl 1) + ip[k + ipp] ;
                 k1 := (k shl 1) + ip[j + ipp] ;
                 xr := a[j1 + aip] ;
                 xi := a[j1 + 1 + aip] ;
                 a[j1 + aip] := a[k1 + aip] ;
                 a[j1 + 1 + aip] := a[k1 + 1 + aip] ;
                 a[k1 + aip] := xr ;
                 a[k1 + 1 + aip] := xi ;
                 Inc(j1,m2) ;
                 Inc(k1,m2) ;
                 xr := a[j1 + aip] ;
                 xi := a[j1 + 1 + aip] ;
                 a[j1 + aip] := a[k1 + aip] ;
                 a[j1 + 1 + aip] := a[k1 + 1 + aip] ;
                 a[k1 + aip] := xr ;
                 a[k1 + 1 + aip] := xi ;
             end ;
         end ;
     end ;

end { TFFT.bitrv2 } ;

function TFFT.CalcCorrelation : integer ;

// calculates a correlation array by convolving the current spectral data
// with a known idle signal "fingerprint".

const

     MIN_F = 13	;	 //100 Hz( 100*1024/8000)
     MAX_F = 230 ; //1800 Hz( 1800*1024/8000)
     F_50 = 6	 ;	 //50 Hz( 50*1024/8000) sub-peak max spacing
     F_25 = 3	 ;	 //25 Hz( 25*1024/8000) sub-peak valley spacing

var

     i,j : integer ;
     acc : double ;
     max : double ;
     peak : integer ;

begin

     max := -1e-100 ;
     peak := 0 ; // PW added !

// create an array of correlated data in across the frequency spectrum
// looking for a "frequency fingerprint" of the IDLE BPSK data string

     for i := MIN_F to MAX_F-1 do begin

         acc := 0.0 ;
         for j := 0 to 200-1 do begin
             acc := acc + CORREL_ARRAY[j] * pFFTAveBuf[(i+j) shl 1] ;
         end ;
         CoreData[i] := acc ;
         if (acc > max) then begin
            max := acc ;
            peak := i ;
         end ;
     end ;

//qualify the correlated data by making sure it has a secondary peak
//  50 Hz away from the main peak

     max := 2.0*CoreData[peak]-CoreData[peak+F_25]-CoreData[peak-F_25] ;
     acc := 2.0*CoreData[peak]-CoreData[peak+F_50]-CoreData[peak-F_50] ;

     if (max > 2.3*acc) then begin
         peak := ((peak+CORR_SIZE div 2)*8000) div 1024 ;
     end
     else begin
         peak := -1 ;
     end ;

     result := peak ;

end { TFFT.CalcCorrelation } ;

procedure TFFT.CalcFFT(var InBuf: TFFTBuffer);

var

     i : integer ;

begin

//   "InBuf[]" is first multiplied by a window function and then
//   calculates an "FFT_SIZE" point FFT on "InBuf[]".
//   The result is converted to dB or 4th root and stored in pFFTAveBuf
//   If "Ave" is > 1, a LP smoothing filter
//   is calculated on the output.

     bOverload := false ;
     for i := 0 to FFT_SIZE-1 do begin
         if (InBuf[i] > 32768.0*0.90) then begin // flag overload if within 10% of max
             bOverload := true ;
         end ;
         pFFTInBuf[i] := WindowTbl[i] * InBuf[i] ; //window the data
     end ;

//   Calculate the FFT

     bitrv2(FFT_SIZE, WorkArea, 2, pFFTInBuf,0) ;
     cftfsub(FFT_SIZE, pFFTInBuf, SinCosTbl) ;
     rftfsub(FFT_SIZE, pFFTInBuf, WorkArea[1], SinCosTbl , WorkArea[0]);

end { TFFT.CalcFFT } ;

procedure TFFT.cft1st(n: integer; var a: TFFTBuffer; var w: TSinCosTbl);
var
    j, k1, k2 : integer ;
    wk1r, wk1i, wk2r, wk2i, wk3r, wk3i : double ;
    x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i : double ;
begin

     x0r := a[0] + a[2] ;
     x0i := a[1] + a[3] ;
     x1r := a[0] - a[2] ;
     x1i := a[1] - a[3] ;
     x2r := a[4] + a[6] ;
     x2i := a[5] + a[7] ;
     x3r := a[4] - a[6] ;
     x3i := a[5] - a[7] ;
     a[0] := x0r + x2r ;
     a[1] := x0i + x2i ;
     a[4] := x0r - x2r ;
     a[5] := x0i - x2i ;
     a[2] := x1r - x3i ;
     a[3] := x1i + x3r ;
     a[6] := x1r + x3i ;
     a[7] := x1i - x3r ;
     wk1r := w[2] ;
     x0r := a[8] + a[10] ;
     x0i := a[9] + a[11] ;
     x1r := a[8] - a[10] ;
     x1i := a[9] - a[11] ;
     x2r := a[12] + a[14] ;
     x2i := a[13] + a[15] ;
     x3r := a[12] - a[14] ;
     x3i := a[13] - a[15] ;
     a[8] := x0r + x2r ;
     a[9] := x0i + x2i ;
     a[12] := x2i - x0i ;
     a[13] := x0r - x2r ;
     x0r := x1r - x3i ;
     x0i := x1i + x3r ;
     a[10] := wk1r * (x0r - x0i) ;
     a[11] := wk1r * (x0r + x0i) ;
     x0r := x3i + x1r ;
     x0i := x3r - x1i ;
     a[14] := wk1r * (x0i - x0r) ;
     a[15] := wk1r * (x0i + x0r) ;
     k1 := 0 ;
     j := 16 ;
     while (j < n) do begin
         Inc(k1,2) ;
         k2 := k1 shl 1 ;
         wk2r := w[k1] ;
         wk2i := w[k1 + 1] ;
         wk1r := w[k2] ;
         wk1i := w[k2 + 1] ;
         wk3r := wk1r - 2 * wk2i * wk1i ;
         wk3i := 2 * wk2i * wk1r - wk1i ;
         x0r := a[j] + a[j + 2] ;
         x0i := a[j + 1] + a[j + 3] ;
         x1r := a[j] - a[j + 2] ;
         x1i := a[j + 1] - a[j + 3] ;
         x2r := a[j + 4] + a[j + 6] ;
         x2i := a[j + 5] + a[j + 7] ;
         x3r := a[j + 4] - a[j + 6] ;
         x3i := a[j + 5] - a[j + 7] ;
         a[j] := x0r + x2r ;
         a[j + 1] := x0i + x2i ;
         x0r := x0r - x2r ;
         x0i := x0i - x2i;
         a[j + 4] := wk2r * x0r - wk2i * x0i ;
         a[j + 5] := wk2r * x0i + wk2i * x0r ;
         x0r := x1r - x3i ;
         x0i := x1i + x3r ;
         a[j + 2] := wk1r * x0r - wk1i * x0i ;
         a[j + 3] := wk1r * x0i + wk1i * x0r ;
         x0r := x1r + x3i ;
         x0i := x1i - x3r ;
         a[j + 6] := wk3r * x0r - wk3i * x0i ;
         a[j + 7] := wk3r * x0i + wk3i * x0r ;
         wk1r := w[k2 + 2] ;
         wk1i := w[k2 + 3] ;
         wk3r := wk1r - 2 * wk2r * wk1i ;
         wk3i := 2 * wk2r * wk1r - wk1i ;
         x0r := a[j + 8] + a[j + 10] ;
         x0i := a[j + 9] + a[j + 11] ;
         x1r := a[j + 8] - a[j + 10] ;
         x1i := a[j + 9] - a[j + 11] ;
         x2r := a[j + 12] + a[j + 14] ;
         x2i := a[j + 13] + a[j + 15] ;
         x3r := a[j + 12] - a[j + 14] ;
         x3i := a[j + 13] - a[j + 15] ;
         a[j + 8] := x0r + x2r ;
         a[j + 9] := x0i + x2i ;
         x0r := x0r - x2r ;
         x0i := x0i - x2i ;
         a[j + 12] := -wk2i * x0r - wk2r * x0i ;
         a[j + 13] := -wk2i * x0i + wk2r * x0r ;
         x0r := x1r - x3i ;
         x0i := x1i + x3r ;
         a[j + 10] := wk1r * x0r - wk1i * x0i ;
         a[j + 11] := wk1r * x0i + wk1i * x0r ;
         x0r := x1r + x3i ;
         x0i := x1i - x3r ;
         a[j + 14] := wk3r * x0r - wk3i * x0i ;
         a[j + 15] := wk3r * x0i + wk3i * x0r ;
         Inc(j,16) ;
     end ;

end { TFFT.cft1st } ;

procedure TFFT.cftfsub(n: integer; var a: TFFTBuffer; var w: TSinCosTbl);

var

    j, j1, j2, j3, l : integer ;
    x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i : double ;

begin

     l := 2 ;
     if (n > 8) then begin
         cft1st(n, a, w) ;
         l := 8 ;
         while ((l shl 2) < n) do begin
             cftmdl(n, l, a, w) ;
             l := l shl  2 ;
         end ;
     end ;
     if ((l shl 2) = n) then begin
         j := 0 ;
         while (j < l) do begin
            j1 := j + l ;
            j2 := j1 + l ;
            j3 := j2 + l ;
            x0r := a[j] + a[j1] ;
            x0i := a[j + 1] + a[j1 + 1] ;
            x1r := a[j] - a[j1] ;
            x1i := a[j + 1] - a[j1 + 1] ;
            x2r := a[j2] + a[j3] ;
            x2i := a[j2 + 1] + a[j3 + 1] ;
            x3r := a[j2] - a[j3] ;
            x3i := a[j2 + 1] - a[j3 + 1] ;
            a[j] := x0r + x2r ;
            a[j + 1] := x0i + x2i ;
            a[j2] := x0r - x2r ;
            a[j2 + 1] := x0i - x2i ;
            a[j1] := x1r - x3i ;
            a[j1 + 1] := x1i + x3r ;
            a[j3] := x1r + x3i ;
            a[j3 + 1] := x1i - x3r ;
            Inc(j,2) ;
         end ;
     end
     else begin
         j := 0 ;
         while (j < l) do begin
            j1 := j + l ;
            x0r := a[j] - a[j1] ;
            x0i := a[j + 1] - a[j1 + 1] ;
            a[j] := a[j] + a[j1] ;
            a[j + 1] := a[j + 1] +  a[j1 + 1] ;
            a[j1] := x0r ;
            a[j1 + 1] := x0i ;
            Inc(j,2) ;
         end ;
     end ;

end { TFFT.cftfsub } ;

procedure TFFT.cftmdl(n, l: integer; var a: TFFTBuffer; var w: TSinCosTbl);

var

     j, j1, j2, j3, k, k1, k2, m, m2 : integer ;
     wk1r, wk1i, wk2r, wk2i, wk3r, wk3i : double ;
     x0r, x0i, x1r, x1i, x2r, x2i, x3r, x3i : double ;

begin

     m := l shl 2 ;
     j := 0 ;
     while (j < l) do begin
         j1 := j + l ;
         j2 := j1 + l ;
         j3 := j2 + l ;
         x0r := a[j] + a[j1] ;
         x0i := a[j + 1] + a[j1 + 1] ;
         x1r := a[j] - a[j1] ;
         x1i := a[j + 1] - a[j1 + 1] ;
         x2r := a[j2] + a[j3] ;
         x2i := a[j2 + 1] + a[j3 + 1] ;
         x3r := a[j2] - a[j3] ;
         x3i := a[j2 + 1] - a[j3 + 1] ;
         a[j] := x0r + x2r ;
         a[j + 1] := x0i + x2i ;
         a[j2] := x0r - x2r ;
         a[j2 + 1] := x0i - x2i ;
         a[j1] := x1r - x3i ;
         a[j1 + 1] := x1i + x3r ;
         a[j3] := x1r + x3i ;
         a[j3 + 1] := x1i - x3r ;
         Inc(j,2) ;
     end ;
     wk1r := w[2] ;
     j := m ;
     while (j < l + m) do begin
         j1 := j + l ;
         j2 := j1 + l ;
         j3 := j2 + l ;
         x0r := a[j] + a[j1] ;
         x0i := a[j + 1] + a[j1 + 1] ;
         x1r := a[j] - a[j1] ;
         x1i := a[j + 1] - a[j1 + 1] ;
         x2r := a[j2] + a[j3] ;
         x2i := a[j2 + 1] + a[j3 + 1] ;
         x3r := a[j2] - a[j3] ;
         x3i := a[j2 + 1] - a[j3 + 1] ;
         a[j] := x0r + x2r ;
         a[j + 1] := x0i + x2i ;
         a[j2] := x2i - x0i ;
         a[j2 + 1] := x0r - x2r ;
         x0r := x1r - x3i ;
         x0i := x1i + x3r ;
         a[j1] := wk1r * (x0r - x0i) ;
         a[j1 + 1] := wk1r * (x0r + x0i) ;
         x0r := x3i + x1r ;
         x0i := x3r - x1i ;
         a[j3] := wk1r * (x0i - x0r) ;
         a[j3 + 1] := wk1r * (x0i + x0r) ;
         Inc(j,2) ;
     end ;
     k1 := 0 ;
     m2 := m shl 1 ;
     k := m2 ;
     while (k < n) do begin
         Inc(k1,2) ;
         k2 := k1 shl 1 ;
         wk2r := w[k1] ;
         wk2i := w[k1 + 1] ;
         wk1r := w[k2] ;
         wk1i := w[k2 + 1] ;
         wk3r := wk1r - 2 * wk2i * wk1i ;
         wk3i := 2 * wk2i * wk1r - wk1i ;
         j := k ;
         while (j < l + k) do begin
             j1 := j + l ;
             j2 := j1 + l ;
             j3 := j2 + l ;
             x0r := a[j] + a[j1] ;
             x0i := a[j + 1] + a[j1 + 1] ;
             x1r := a[j] - a[j1] ;
             x1i := a[j + 1] - a[j1 + 1] ;
             x2r := a[j2] + a[j3] ;
             x2i := a[j2 + 1] + a[j3 + 1] ;
             x3r := a[j2] - a[j3] ;
             x3i := a[j2 + 1] - a[j3 + 1] ;
             a[j] := x0r + x2r ;
             a[j + 1] := x0i + x2i ;
             x0r := x0r - x2r ;
             x0i := x0i - x2i ;
             a[j2] := wk2r * x0r - wk2i * x0i ;
             a[j2 + 1] := wk2r * x0i + wk2i * x0r ;
             x0r := x1r - x3i ;
             x0i := x1i + x3r ;
             a[j1] := wk1r * x0r - wk1i * x0i ;
             a[j1 + 1] := wk1r * x0i + wk1i * x0r ;
             x0r := x1r + x3i ;
             x0i := x1i - x3r ;
             a[j3] := wk3r * x0r - wk3i * x0i ;
             a[j3 + 1] := wk3r * x0i + wk3i * x0r ;
             Inc(j,2) ;
         end ;
         wk1r := w[k2 + 2] ;
         wk1i := w[k2 + 3] ;
         wk3r := wk1r - 2 * wk2r * wk1i ;
         wk3i := 2 * wk2r * wk1r - wk1i ;
         j := k + m ;
         while (j < l + (k + m)) do begin
             j1 := j + l ;
             j2 := j1 + l ;
             j3 := j2 + l ;
             x0r := a[j] + a[j1] ;
             x0i := a[j + 1] + a[j1 + 1] ;
             x1r := a[j] - a[j1] ;
             x1i := a[j + 1] - a[j1 + 1] ;
             x2r := a[j2] + a[j3] ;
             x2i := a[j2 + 1] + a[j3 + 1] ;
             x3r := a[j2] - a[j3] ;
             x3i := a[j2 + 1] - a[j3 + 1] ;
             a[j] := x0r + x2r ;
             a[j + 1] := x0i + x2i ;
             x0r := x0r - x2r ;
             x0i := x0i - x2i ;
             a[j2] := -wk2i * x0r - wk2r * x0i ;
             a[j2 + 1] := -wk2i * x0i + wk2r * x0r ;
             x0r := x1r - x3i ;
             x0i := x1i + x3r ;
             a[j1] := wk1r * x0r - wk1i * x0i ;
             a[j1 + 1] := wk1r * x0i + wk1i * x0r ;
             x0r := x1r + x3i ;
             x0i := x1i - x3r ;
             a[j3] := wk3r * x0r - wk3i * x0i ;
             a[j3 + 1] := wk3r * x0i + wk3i * x0r ;
             Inc(j,2) ;
         end ;
         Inc(k,m2) ;
     end ;

end { TFFT.cftmdl } ;

constructor TFFT.Create;

var

     i : integer ;

begin

     WorkArea[0] := 0 ;
     makewt(FFT_SIZE div 4, WorkArea, SinCosTbl,0) ;
     makect(FFT_SIZE div 4, WorkArea, SinCosTbl, WorkArea[0]);

     for i := 0 to FFT_SIZE-1  do	begin
         pFFTAveBuf[i] := 0.0 ;
         WindowTbl[i] := (0.5 - 0.5*Cos( (K_2PI*i)/(FFT_SIZE-1) )) ; //Hanning
     end ;

     iAveSize := 10 ;
     iLastAve := 10 ;
     bLogMode := true ;
     bLastLogMode := true ;
     dGain := 10.0;
     dClip := 0.0;
     bOverload := false ;

end ; { TFFT.Create }

function TFFT.GetFFTData(iStart, iStop: integer; var pOutBuf:TFFTOutBuf) : boolean ;

//	the range "start" to "stop" is multiplied by "gain" and copied
//  into "OutBuf[]".
// The function returns TRUE if the input is overloaded

var
     i : integer ;

begin
     for i := iStart to iStop do begin	//copy and scale into OutBuf[]
//       pOutBuf[i] := Trunc(dGain*pFFTAveBuf[i shl 1]) ;
         pOutBuf[i] := Trunc(dGain*pFFTAveBuf[i shl 1]) ;
     end ;
     result := bOverload ;
end { TFFT.GetFFTData } ;

function TFFT.GetFFTPeak(iCenter, iRange: integer): integer;

// returns the nearest maximum frequency peak around the center freq
//  and +/- the range frequency.

var

     i : integer ;

begin

     for i := 0 to 200-1 do begin
         CoreData[i] := pFFTAveBuf[(106+i) shl 1] ;
     end ;

     result := iCenter + iRange ;

end { TFFT.GetFFTPeak } ;

procedure TFFT.makect(nc:integer; var ip:TWorkArea; var c:TSinCosTbl;cip:integer { index into c } ) ;

var

    nch, j : integer ;
    delta : double ;

begin

    ip[1] := nc ;
    if (nc > 1) then begin
        nch := nc shr 1 ;
        delta := ArcTan(1.0) / nch ;
        c[0+cip] := Cos(delta * nch) ;
        c[nch+cip] := 0.5 * c[0+cip] ;
        for j := 1 to nch-1 do begin
            c[j+cip] := 0.5 * Cos(delta * j) ;
            c[nc - j+cip] := 0.5 * Sin(delta * j) ;
        end ;
    end ;

end { TFFT.makect } ;

procedure TFFT.makewt(nw:integer; var ip:TWorkArea; var w:TSinCosTbl;wip:integer { index into w } ) ;

var

     nwh, j : integer ;
     delta, x, y : double ;

begin

    ip[0] := nw ;
    ip[1] := 1 ;
    if (nw > 2) then begin
        nwh := nw shr 1 ;
        delta := ArcTan(1.0) / nwh ;
        w[0+wip] := 1 ;
        w[1+wip] := 0 ;
        w[nwh+wip] := Cos(delta * nwh) ;
        w[nwh + 1 +wip] := w[nwh+wip] ;
        j := 2 ;
        while (j < nwh) do begin
            x := Cos(delta * j) ;
            y := Sin(delta * j) ;
            w[j+wip] := x ;
            w[j + 1+wip] := y ;
            w[nw - j+wip] := y ;
            w[nw - j + 1+wip] := x ;
            Inc(j,2) ;
        end ;
        bitrv2(nw, ip,  2, w,wip);
    end ;

end { TFFT.makewt } ;

procedure TFFT.ResetFFT;
var
     i : integer ;
begin
     for i := 0 to FFT_SIZE-1 do begin
          pFFTAveBuf[i] := 0.0 ;
     end ;
end { TFFT.ResetFFT } ;

procedure TFFT.rftfsub(n:integer; var a:TFFTBuffer; nc:integer; var c:TSinCosTbl;cpp:integer { index into c}) ;

var

    tmp : double ;
    j, k, kk, ks, m : integer ;
    wkr, wki, xr, xi, yr, yi : double ;

begin

     ks := (nc shl 2) div n ;
     kk := 0 ;
     m := n shr 1 ;
     if (bLogMode) then begin
         k := 2 ;
         while (k < m) do begin

             j := n - k ;
             Inc(kk,ks) ;
             wkr := 0.5 - c[nc - kk +cpp] ;
             wki := c[kk+cpp] ;
             xr := a[k] - a[j] ;
             xi := a[k + 1] + a[j + 1] ;
             yr := wkr * xr - wki * xi ;
             yi := wkr * xi + wki * xr ;
             a[k] := a[k]- yr ;
             xi := a[k]*a[k];
             a[k+1] := a[k+1] - yi ;
             xi := xi + (a[k+1]*a[k+1]) ;
             a[j] := a[j] + yr ;
             xr := a[j]*a[j] ;
             a[j+1] := a[j+1] - yi ;
             xr := xr + (a[j+1]*a[j+1]) ;
             tmp := Log10(xi+K_C) + K_B ;
             tmp := tmp - dClip ;
             if (tmp<0.0) then begin
                 pFFTAveBuf[k] := (1.0-1.0/iAveSize)*pFFTAveBuf[k] ;
             end
             else begin
                 pFFTAveBuf[k] := (1.0-1.0/iAveSize)*pFFTAveBuf[k] + (1.0/iAveSize)*tmp ;
             end ;
             tmp := log10(xr+K_C) + K_B ;
             tmp := tmp - dClip ;
             if (tmp<0.0) then begin
                pFFTAveBuf[j] := (1.0-1.0/iAveSize)*pFFTAveBuf[j] ;
             end
             else begin
                pFFTAveBuf[j] := (1.0-1.0/iAveSize)*pFFTAveBuf[j] +	(1.0/iAveSize)*tmp ;
             end ;
             Inc(k,2) ;
         end ;
     end
	   else begin
         k := 2 ;
         while (k < m) do begin

             j := n - k ;
             kk := kk + ks ;
             wkr := 0.5 - c[nc - kk+cpp] ;
             wki := c[kk+cpp] ;
             xr := a[k] - a[j] ;
             xi := a[k + 1] + a[j + 1] ;
             yr := wkr * xr - wki * xi ;
             yi := wkr * xi + wki * xr ;
             a[k] := a[k] - yr ;
             xi := a[k]*a[k] ;
             a[k+1] := a[k+1] - yi ;
             xi := xi + ( a[k+1]*a[k+1]) ;
             a[j] := a[j] + yr ;
             xr := a[j]*a[j] ;
             a[j+1] := a[j+1] - yi ;
             xr := xr + (a[j+1]*a[j+1]) ;

             pFFTAveBuf[k] := (1.0-1.0/iAveSize)*pFFTAveBuf[k] +
                              (1.0/iAveSize)*Power(xi , K_ROOT)/K_ROOTGN ;
             pFFTAveBuf[j] := (1.0-1.0/iAveSize)*pFFTAveBuf[j] +
                              (1.0/iAveSize)*Power(xr , K_ROOT)/K_ROOTGN ;
             Inc(k,2) ;
         end ;
     end ;
     pFFTAveBuf[FFT_SIZE div 2] := pFFTAveBuf[FFT_SIZE div 2 - 2];

end { TFFT.rftfsub } ;

procedure TFFT.SetFFTParams(iAve: integer; Gain: double; iType: integer);
begin

     bLogMode := iType<>0 ;

     if ((iType>=10) and (iType<=99)) then begin
         dClip := iType / 10.0 ;
     end
     else begin
         dClip := 0.0 ;
     end ;

     if (iAve>0) then begin
         iAveSize := iAve ;
     end
     else begin
         iAveSize := 1 ;
     end ;

     if ((iLastAve <> iAve) or (bLastLogMode <> bLogMode)) then begin
         ResetFFT ;
     end ;

     bLastLogMode := bLogMode ;
     iLastAve := iAveSize ;
     dGain := 0.1*(Gain*10.0/(10.0-dClip)) ;

end { TFFT.SetFFTParams } ;

end.
