(*
                  umdlIOCntrl
       (c) 2003 Paul C. L. Willmott, VP9MU

          Based on code from AO40RCV
         (c) 2000 Moe Wheatley AE4JY
             ae4jy@mindspring.com

 This program is free software; you can redistribute it
 and/or modify it under the terms of the GNU General
 Public License as published by the Free Software
 Foundation; either version 2 of the License, or at
 your option, any later version.

 This program is distributed in the hope that it will
 be useful, but WITHOUT ANY WARRANTY; without even the
 implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE.  See the GNU General Public
 License for more details.

 You should have received a copy of the GNU General
 Public License along with this program; if not, write
 to the Free Software Foundation, Inc., 59 Temple
 Place, Suite 330, Boston, MA  02111-1307  USA

            Contact : vp9mu@amsat.org
*)

(*
     Revision History
     ----------------

     1.0.0      Created July 21, 2003 PCLW
*)


unit umdlIOCntrl ;

interface

uses

   Classes,
   umdlSound,
   umdlFFT,
   MMSystem,
   SyncObjs,
   umdlBPSKDet,
   Messages,
   Windows ;

const

   MAXTIMER = 10 ;	//wait for 1 seconds for thread to start

type

   TIOCntrl = class(TThread)
   private

//   Variables

     iProgressCount   : integer          ; // No Idea what this does yet!
     iProgress        : integer          ; // No Idea what this does yet!

     bThreadActive    : boolean          ; // True if thread active
     bBufferRdy       : boolean          ;
     sErrorMessage    : string           ; // Human readble error message
     iDLLStatus       : integer          ; // Main Status for calling program
     iDLLErrStatus    : integer          ;
     hWnd             : HWND             ; // Handle of Form to post messages to
     pBuffer          : TFFTWaveBuffer   ; // FFT Input Buffer, PCM converted to Double
     pSoundDev        : TSound           ; // Soundcard Driver Class
     pFFT             : TFFT             ; // FFT Class
     pWaveInFormatEx  : TWAVEFORMATEX    ; // Wave In Format
     pCriticalSection : TCriticalSection ; // for keeping threads from stomping on each other

//   Functions

     procedure StatusChange(iCode:integer) ;
     procedure StopThread ;
     procedure ProcessError(iErrorCode:integer ) ;

   protected

     procedure Execute ; override ;

   public

//   Publicly exposed Variables

     iCardNum : integer  ; // Input Soundcard Number
//   pBPSKDet : TBPSKDet ; // PSK Detector Class

//   Methods

     function    GetErrorMsg : string ;
     function    GetFFTData(var pData:TFFTOutBuf; iStart, iEnd:integer): boolean ;
//   procedure   SetFFTParams(iAve : integer; dGain:double; iType:integer) ;
     procedure   StopIO ;
     function    StartIO(Handle:HWND): integer ;
     constructor Create(CreateSuspended:boolean) ;
     destructor  Destroy ;

   end ;

var

   IOCntrl : TIOCntrl ; // Only 1 instance

implementation

uses

   umdlErrorCodes,
   SysUtils ,
   umdlBPSK400 ;

{ TIOCntrl }

constructor TIOCntrl.Create(CreateSuspended:boolean);
begin

//   Give normal priority, and don't destroy class when thread ends

     inherited ;
     Priority := tpNormal ;
     FreeOnTerminate := false ;

//   Set Status Flags

     bThreadActive := false ;
     iDLLStatus := DLLSTAT_OFF ;

//   Setup Soundcard values

     pWaveInFormatEx.wFormatTag      := WAVE_FORMAT_PCM ;
     pWaveInFormatEx.nChannels       := WAVE_NUM_CHANNELS ;
     pWaveInFormatEx.wBitsPerSample  := WAVE_BITS_PER_SAMPLE ;
     pWaveInFormatEx.nSamplesPerSec  := WAVE_SAMPLE_RATE ;
     pWaveInFormatEx.nBlockAlign     := pWaveInFormatEx.nChannels * (pWaveInFormatEx.wBitsPerSample div 8) ;
     pWaveInFormatEx.nAvgBytesPerSec := pWaveInFormatEx.nSamplesPerSec * pWaveInFormatEx.nBlockAlign ;
     pWaveInFormatEx.cbSize          := 0 ;

//   Create Soundcard driver instance

     pSoundDev := TSound.Create ;

//   Create FFT instance

     pFFT := TFFT.Create ;
     pFFT.ResetFFT ;
     pFFT.SetFFTParams(5,255,20);
     pFFT.ResetFFT ;

//   Create PSK Detector instance

     pBPSKDet := TBPSKDet.Create ;

//   Create Critical section for program interface

     pCriticalSection := TCriticalSection.Create ;

end { TIOCntrl.Create } ;

destructor TIOCntrl.Destroy;
begin
     inherited ;
     pCriticalSection.Destroy ;
     pBPSKDet.Destroy ;
     pFFT.Destroy ;
     pSoundDev.Destroy ;
end { TIOCntrl.Destroy } ;

procedure TIOCntrl.Execute;

//   Main I/O thread processing loop
//   Process for Receiving audio data from soundcard

var

     iResult    : integer ;
     iErrorCode : integer ;

begin

     inherited;

     Sleep(50) ;

     bBufferRdy := false ;

     iErrorCode := pSoundDev.InOpen(pWaveInFormatEx, 0, iCardNum) ;
     if (iErrorCode <> NO_ERRORS) then begin
         ProcessError(iErrorCode) ;
         StopThread ;
         Exit ;
     end ;

     pBPSKDet.Init(WAVE_SAMPLE_RATE, WAVE_SAMPLE_SIZE, hWnd) ;

     bThreadActive := true ;
     StatusChange(DLLSTAT_RX_NO_SIGNAL) ;
     Sleep(500) ;

     iProgressCount := 0 ;
     while (not Terminated) do begin

         iResult := pSoundDev.InRead(pBuffer, WAVE_SAMPLE_SIZE) ;
         if (iResult <> WAVE_SAMPLE_SIZE) then begin

             iErrorCode := pSoundDev.GetError ;

             if ((iErrorCode = SOUNDIN_ERR_OVERFLOW) or
                 (iErrorCode = SOUNDIN_ERR_TIMEOUT)) then begin
//   cpu couldn't keep up so try again
                  pSoundDev.InClose ;
                  StatusChange(DLLSTAT_CPUSLOW) ;
                  iErrorCode := pSoundDev.InOpen(pWaveInFormatEx, 0, iCardNum) ;
             end ;

             if (iErrorCode <> NO_ERRORS) then begin //must be some other soundcard error
                 pSoundDev.InClose ;
                 ProcessError(iErrorCode) ;
                 StopThread ;
                 Exit ;
             end ;

         end
         else begin
             iProgress := 0 ;
         end ;

//   Calculate 1024 pt. FFT on both halves of the 2048 sample buffer

         pCriticalSection.Acquire ;
         pFFT.CalcFFT(TFFTWaveBuffer(pBuffer).pBuffer1) ;
         pFFT.CalcFFT(TFFTWaveBuffer(pBuffer).pBuffer2) ;
         bBufferRdy := true ;
         pCriticalSection.Leave ;

         PostMessage(hWnd, MSG_FFTDATARDY, iProgress, 0);

// Process detector.  Returns decoder state

         iResult := pBPSKDet.ProcBPSKDet(pBuffer) ;
         StatusChange(iResult) ;

         if (pBPSKDet.OkToCheckFreq) then begin
             pBPSKDet.SetRXFrequency(pFFT.CalcCorrelation,false);
         end ;

     end ;

     StatusChange(DLLSTAT_OFF);
     pSoundDev.InClose ;
     bThreadActive := false ;

end { TIOCntrl.Execute } ;

function TIOCntrl.GetErrorMsg : string ;
begin
     result := sErrorMessage ;
end { TIOCntrl.GetErrorMsg } ;

function TIOCntrl.GetFFTData(var pData: TFFTOutBuf; iStart,
  iEnd: integer): boolean;

///////////////////////////////////////////////////////////////////////////
// Gets latest raw data and put's it in users array
///////////////////////////////////////////////////////////////////////////

begin

     result := false ;
     Fillchar(pData,SizeOf(pData),0) ;
     if (not Terminated and bBufferRdy) then begin
         pCriticalSection.Acquire ;
         result := pFFT.GetFFTData(iStart, iEnd, pData) ;
         pCriticalSection.Leave ;
     end ;
     bBufferRdy := false ;
     result := true ;

end { TIOCntrl.GetFFTData } ;

procedure TIOCntrl.ProcessError(iErrorCode:integer) ;
begin

     iDLLErrStatus := iErrorCode ;

     case iErrorCode of
     NO_ERRORS : begin
         sErrorMessage := '' ;
     end ;
//   MEMORY_ERROR:
     MMSYSERR_NOMEM: begin
         sErrorMessage := 'Memory Allocation error' ;
         iErrorCode := DLLERR_MEMORY ;
     end ;
     SOUNDIN_ERR_NOTOPEN: begin
         sErrorMessage := 'Tried to read and soundcard is not open' ;
         iErrorCode := DLLERR_IN_NOTOPEN ;
     end ;
     SOUNDIN_ERR_OVERFLOW: begin
         sErrorMessage := 'Input buffers overflowed' ;
         iErrorCode := DLLERR_IN_OVERFLOW ;
     end ;
     SOUNDIN_ERR_TIMEOUT: begin
         sErrorMessage := 'Timed out waiting for input buffers' ;
         iErrorCode := DLLERR_IN_TIMEOUT ;
     end ;
     SOUNDOUT_ERR_NOTOPEN: begin
         sErrorMessage := 'Tried to write and soundcard is not open' ;
         iErrorCode := DLLERR_OUT_NOTOPEN ;
     end ;
     SOUNDOUT_ERR_UNDERFLOW: begin
         sErrorMessage := 'Output buffers Underflowed' ;
         iErrorCode := DLLERR_OUT_UNDERFLOW ;
     end ;
     SOUNDOUT_ERR_TIMEOUT: begin
         sErrorMessage := 'Timed out waiting for output buffers' ;
         iErrorCode := DLLERR_OUT_TIMEOUT ;
     end ;
     WAVERR_BADFORMAT: begin
         sErrorMessage := 'Wave file is not 8000Hz, Mono format' ;
         iErrorCode := DLLERR_WAV_BADFORMAT ;
     end ;
     WAVERR_STILLPLAYING: begin
         sErrorMessage := 'still something playing' ;
         iErrorCode := DLLERR_WAV_STILLPLAYING ;
     end ;
     WAVERR_UNPREPARED: begin
         sErrorMessage := 'header not prepared ' ;
         iErrorCode := DLLERR_WAV_UNPREPARED ;
     end ;
     WAVERR_SYNC: begin
         sErrorMessage := 'device is synchronous' ;
         iErrorCode := DLLERR_WAV_SYNC ;
     end ;
     MMSYSERR_BADDEVICEID: begin
         sErrorMessage := 'Bad Device ID! Is Soundcard Present?' ;
         iErrorCode := DLLERR_SYS_BADDEVICEID ;
     end ;
     MMSYSERR_NOTENABLED: begin
         sErrorMessage := 'Driver failed enable' ;
         iErrorCode := DLLERR_SYS_NOTENABLED ;
     end ;
     MMSYSERR_ALLOCATED: begin
         sErrorMessage := 'Device already allocated' ;
         iErrorCode := DLLERR_SYS_ALLOCATED ;
     end ;
     MMSYSERR_INVALHANDLE: begin
         sErrorMessage := 'Device handle is invalid' ;
         iErrorCode := DLLERR_SYS_INVALHANDLE ;
     end ;
     MMSYSERR_NODRIVER: begin
         sErrorMessage := 'No device driver present' ;
         iErrorCode := DLLERR_SYS_NODRIVER ;
     end ;
     MMSYSERR_NOTSUPPORTED: begin
         sErrorMessage := 'Function isn''t supported' ;
         iErrorCode := DLLERR_SYS_NOTSUPPORTED ;
     end ;
     MMSYSERR_BADERRNUM: begin
         sErrorMessage := 'Error value out of range' ;
         iErrorCode := DLLERR_SYS_BADERRNUM ;
     end ;
     MMSYSERR_INVALFLAG: begin
         sErrorMessage := 'Invalid flag passed' ;
         iErrorCode := DLLERR_SYS_INVALFLAG ;
     end ;
     MMSYSERR_INVALPARAM: begin
         sErrorMessage := 'Invalid parameter passed' ;
         iErrorCode := DLLERR_SYS_INVALPARAM ;
     end ;
     MMSYSERR_HANDLEBUSY: begin
         sErrorMessage := 'Handle being used' ;
         iErrorCode := DLLERR_SYS_HANDLEBUSY ;
     end ;
     MMSYSERR_NODRIVERCB: begin
         sErrorMessage := 'driver does not call DriverCallback' ;
         iErrorCode := DLLERR_SYS_NODRIVERCB ;
     end ;
(*
     MMSYSERR_INVALIDALIAS:
     MMSYSERR_BADDB:
     MMSYSERR_KEYNOTFOUND:
     MMSYSERR_READERROR:
     MMSYSERR_WRITEERROR:
     MMSYSERR_DELETEERROR:
*)
     MMSYSERR_VALNOTFOUND: begin
         sErrorMessage := 'Registry error' ;
         iErrorCode := DLLERR_SYS_REGISTRY ;
     end ;
     else
         sErrorMessage := 'Unknown Error' ;
         iErrorCode := DLLERR_UNKNOWN ;
     end ;

     iDLLStatus := iErrorCode ;

     if (bThreadActive) then begin
         StatusChange(iErrorCode) ;
     end ;

end;

(*
procedure TIOCntrl.SetFFTParams(iAve: integer; dGain: double;
  iType: integer);
// Initializes the FFT
begin
     pFFT.SetFFTParams(iAve, dGain, iType) ;
end { TIOCntrl.SetFFTParams } ;
*)
function TIOCntrl.StartIO(Handle:HWND): integer;

///////////////////////////////////////////////////////////////////////////
//   Called to start the I/O thread running
///////////////////////////////////////////////////////////////////////////

var

     Timer : integer ;

begin

     hWnd := Handle ;

     Resume ;	//let er rip

     Timer := MAXTIMER;
     while ((iDLLStatus = DLLSTAT_OFF) and (not Terminated) and (Timer>0)) do begin
         Sleep(100) ;
         Dec(Timer) ;
     end ;

     Sleep(100) ;

     if (iDLLStatus >= 0) then begin
        result := iDLLStatus ;
     end
     else begin
        if (iDLLErrStatus > 0) then begin
            result := iDLLErrStatus ;
        end
        else begin
            result := DLLERR_NONE ;
        end ;
     end ;

end { StartIO } ;

procedure TIOCntrl.StatusChange(iCode:integer) ;
begin
     if (iCode <> iDLLStatus) then begin
         iDLLStatus := iCode ;
         PostMessage(hWnd, MSG_STATUSCHANGE, iCode, iProgress) ;
     end ;
end { TIOCntrl.StatusChange } ;

procedure TIOCntrl.StopIO;
//  Stop process thread proceedure and wait for it to finish
//  Called by main program thread to signal data processing thread to stop
//	and then wait for it to stop before returning.
begin
     bThreadActive := false ;
     Terminate ;
     while (not Terminated) do begin
         Sleep(10) ;
     end ;
     iDLLStatus := DLLSTAT_OFF ;
end;

procedure TIOCntrl.StopThread;
begin
     StatusChange(DLLSTAT_OFF);
     iDLLStatus := DLLSTAT_OFF ;
     Terminate ;
end { TIOCntrl.StopThread } ;

end.
