(*
                  umdlSound
       (c) 2003 Paul C. L. Willmott, VP9MU

          Based on code from AO40RCV
         (c) 2000 Moe Wheatley AE4JY
             ae4jy@mindspring.com

 This program is free software; you can redistribute it
 and/or modify it under the terms of the GNU General
 Public License as published by the Free Software
 Foundation; either version 2 of the License, or at
 your option, any later version.

 This program is distributed in the hope that it will
 be useful, but WITHOUT ANY WARRANTY; without even the
 implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE.  See the GNU General Public
 License for more details.

 You should have received a copy of the GNU General
 Public License along with this program; if not, write
 to the Free Software Foundation, Inc., 59 Temple
 Place, Suite 330, Boston, MA  02111-1307  USA

            Contact : vp9mu@amsat.org
*)

(*
     Revision History
     ----------------

     1.0.0      Created July 21, 2003 PCLW
*)

unit umdlSound;

interface

uses

   MMSystem      ,
   SyncObjs      ,
   SysUtils      ,
   umdlErrorCodes,
   umdlFFT       ,
   Classes       ;

const

   NUM_INPUT_SNDBUFFERS = 3    ; // number of sound input buffers to allocate
   WAVE_NUM_CHANNELS    = 1    ; // Mono
   WAVE_BITS_PER_SAMPLE = 16   ; // Sample Depth

   WAVE_BUF_SIZE = WAVE_SAMPLE_SIZE * WAVE_NUM_CHANNELS
                   { * WAVE_BITS_PER_SAMPLE div 8 } ; // Assumes 16 Bit !

type

   TWaveBuffer = packed array[0..WAVE_BUF_SIZE -1] of SmallInt ; // Assumes 16 Bit !

   TFFTWaveBuffer = packed record
      case integer of
      1 : (pBuffer1 : TFFTBuffer ;
           pBuffer2 : TFFTBuffer) ;
      2 : (pBuffer  : packed array[0..FFT_SIZE * 2 -1] of double) ;
   end ;

   TSound = class(TObject)
   private

      bInputOpen        : boolean       ;
      iErrorCode        : cardinal      ;
      iInBufPosition    : integer       ;
      iInSampleLimit    : cardinal      ;
      iInSamplesRead    : cardinal      ;
      hWaveIn           : HWAVEIN       ;
      pInFormat	        : PWAVEFORMATEX ;

      pWaveInHeader : array[0..NUM_INPUT_SNDBUFFERS-1] of TWaveHdr    ;
      pWaveInBuffer : array[0..NUM_INPUT_SNDBUFFERS-1] of TWaveBuffer ;

   public

// Public variables for use by callback functions only.

      pInEvent         : TEvent           ;
      bInWaiting       : boolean          ;
      bInOverflow      : boolean          ;
      iInHeadPtr       : integer          ;
      iInTailPtr       : integer          ;
      pCriticalSection : TCriticalSection ;

// Public member functions for object's users.

      function    InOpen(var pWFX:TWAVEFORMATEX; SampleLimit:cardinal; Card:integer) : cardinal ;
      function    InRead(var pData:TFFTWaveBuffer; Length:integer) : integer ;
      procedure   InClose ;
      function    GetError : cardinal ;
      constructor Create ;
      destructor  Destroy ;

   end ;

implementation

// local defines

const TIMELIMIT = 2000 ;			// time limit to wait on a new soundcard buffer

procedure WaveInCallback(hWaveIn:HWAVEIN; uMsg:cardinal; pSound:TSound; dwParam1, dwParam2:cardinal) stdcall ;

(*
  callback Routine called by mmsystem when a buffer becomes full.
      no system calls except SetEvent may be called from here.
     ( not a method function of TSound Class )
*)

begin

     if (uMsg = MM_WIM_DATA) then begin	// ignore all but buffer full messages

         pSound.pCriticalSection.Acquire ;

         Inc(pSound.iInHeadPtr) ;
         if (pSound.iInHeadPtr >= NUM_INPUT_SNDBUFFERS) then begin // inc ptr
            pSound.iInHeadPtr := 0 ; // handle wrap around
         end ;

         if (pSound.iInHeadPtr = pSound.iInTailPtr)	then begin // check for overflow
             pSound.bInOverflow := true ;
         end ;

         if (pSound.bInWaiting) then begin // if user thread is waiting for buffer
             pSound.bInWaiting := false ;
             pSound.pInEvent.SetEvent ; //signal it
         end ;

         pSound.pCriticalSection.Leave ;

     end ;

end { WaveInCallback } ;

{ TSound }

constructor TSound.Create;
begin

//   Initialise Flags

     bInputOpen  := false ;
     bInWaiting  := false ;
     bInOverflow := false ;

//   Clear Handles and Error Codes

     hWaveIn    := 0 ;
     iErrorCode := 0 ;

//   Critical ection to stop threads stomping on each other

     pCriticalSection := TCriticalSection.Create ;

//   event for callback function to notify new buffer available

     pInEvent := TEvent.Create(nil, false,false,'') ;

end { Create } ;

destructor TSound.Destroy;
begin
     inherited ;
     pInEvent.Destroy ;
     pCriticalSection.Destroy ;
end { Destroy } ;

function TSound.GetError:cardinal;
begin
     result := iErrorCode ;
end { TSound.GetError } ;

procedure TSound.InClose;

//  Closes the Soundcard Input if open

var

     i : integer ;

begin

     bInputOpen := false ;

     if (hWaveIn <> 0) then begin

         waveInReset(hWaveIn) ;
         Sleep(100) ;

         for i := 0 to NUM_INPUT_SNDBUFFERS-1 do begin
             if (pWaveInHeader[i].dwFlags and WHDR_PREPARED <> 0) then begin
                waveInUnprepareHeader(hWaveIn, @pWaveInHeader[i],SizeOf(pWaveInHeader[i])) ;
             end ;
         end ; // for

         Sleep(50) ;

         waveInClose(hWaveIn) ;

         Sleep(50) ;

         hWaveIn := 0 ;

     end ;

end { TSound.InClose } ;

function TSound.InOpen(var pWFX:tWAVEFORMATEX; SampleLimit:cardinal; Card:integer) : cardinal ;

var

     i : integer ;

begin

//   Reset Flags

     bInWaiting  := false ;
     bInOverflow := false ;

//   Reset Pointers and Counts

     iInHeadPtr     := 0           ;
     iInTailPtr     := 0           ;
     iInSamplesRead := 0           ;
     iInBufPosition := 0           ;
     iInSampleLimit := SampleLimit ;

//   Store pointer to wave format for later use

     pInFormat := @pWFX ;

//   open sound card input and get handle(hwvin) to device

     iErrorCode := waveInOpen(@hWaveIn, Card, pInFormat, cardinal(@WaveInCallback), cardinal(Self), CALLBACK_FUNCTION) ;

     if (iErrorCode <> MMSYSERR_NOERROR) then begin
         InClose ;
         result := iErrorCode ;
         Exit ;
     end ;

//   initialise input data buffers

     for i := 0  to NUM_INPUT_SNDBUFFERS-1 do begin

         pWaveInHeader[i].dwBufferLength := SizeOf(pWaveInBuffer[i]) ;
         pWaveInHeader[i].dwFlags := 0 ;
         pWaveInHeader[i].dwUser  := 0 ;
         pWaveInHeader[i].dwBytesRecorded := 0 ;
         pWaveInHeader[i].lpData := PCHAR(@pWaveInBuffer[i]) ;

         iErrorCode := waveInPrepareHeader(hWaveIn, PWaveHdr(@pWaveInHeader[i]), Sizeof(pWaveInHeader[i])) ;

         if (iErrorCode <> MMSYSERR_NOERROR) then begin
             InClose ;
             result :=  iErrorCode ;
             Exit ;
         end ;

         iErrorCode := waveInAddBuffer(hWaveIn, PWaveHdr(@pWaveInHeader[i]), SizeOf(pwaveInHeader[i])) ;
         if (iErrorCode <> MMSYSERR_NOERROR) then begin
             InClose ;
             result := iErrorCode ;
         end ;

     end ; // for

//   start input capturing to buffer

     iErrorCode := waveInStart(hWaveIn) ;
     if (iErrorCode <> MMSYSERR_NOERROR) then begin
         InClose ;
         result := iErrorCode ;
         Exit ;
     end ;

     bInputOpen := true ;
     result := 0 ;

end { TSound.InOpen } ;

function TSound.InRead(var pData:TFFTWaveBuffer; Length:integer) : integer ;

(*
 Reads 'Length' samples of double data from the opened soundcard input

	returns:

   Length	if 'Length' samples were succesfully read
   0 =	if reaches the specified sample limit
	-1 =	if there is an error ( use GetError() to retrieve error )
*)

var

     i : integer ;

begin

//   Check if sound card input open

     if (not bInputOpen) then begin
         iErrorCode := SOUNDIN_ERR_NOTOPEN ;
         result := -1 ;
         Exit ;
     end ;

//   Check if in overflow

     if (bInOverflow) then begin
         iErrorCode := SOUNDIN_ERR_OVERFLOW ;
         InClose ;
         result := -1 ;
         Exit ;
     end ;

//   Lock out other threads

     pCriticalSection.Acquire ;

     if (iInTailPtr = iInHeadPtr) then begin	//if no buffer is filled

//   wait for mmsystem to fill up a new buffer

         bInWaiting := true ;

         pCriticalSection.Leave ;

         if (pInEvent.WaitFor(TIMELIMIT) <> wrSignaled) then begin
             iErrorCode := SOUNDIN_ERR_TIMEOUT ; // if took too long error out
             InClose ;
             result := -1 ;
             Exit ;
         end ;

     end
     else begin
         pCriticalSection.Leave ;
     end ;

//   here if there is data to retrieved

     for i := 0 to Length-1 do begin
         pData.pBuffer[i] := pWaveInBuffer[iInTailPtr,iInBufPosition] ;
         Inc(iInBufPosition) ;
     end ; // for

     if (iInBufPosition >= WAVE_BUF_SIZE) then begin		//finished with this buffer
                                                        //so add it back in to the Queue
         pWaveInHeader[iInTailPtr].dwBytesRecorded := 0 ;
         waveInAddBuffer(hWaveIn, @pWaveInHeader[iInTailPtr], SizeOf(pWaveInHeader[iInTailPtr])) ;
         iInBufPosition := 0 ;
         Inc(iInTailPtr) ;
         if (iInTailPtr >= NUM_INPUT_SNDBUFFERS) then begin // handle wrap around
             iInTailPtr := 0 ;
         end ;

         Inc(iInSamplesRead, Length) ;
         if ((iInSampleLimit <> 0) and (iInSamplesRead >= iInSampleLimit)) then begin
             InClose ;
             result := 0 ;
             Exit ;
         end ;

     end ;

     result := Length

end { TSound.InRead } ;

end.
